"""Standard packed message format class for DIDComm V2."""

import logging

try:
    from didcomm_messaging import DIDCommMessaging
    from didcomm_messaging.crypto.backend.askar import CryptoServiceError
except ImportError as err:
    raise ImportError("Install the didcommv2 extra to use this module.") from err

import json
from typing import Sequence, Tuple, Union

from ..core.profile import ProfileSession
from ..messaging.base_message import DIDCommVersion
from ..messaging.util import time_now
from ..wallet.base import BaseWallet
from ..wallet.error import WalletNotFoundError
from .error import WireFormatParseError
from .inbound.receipt import MessageReceipt
from .wire_format import BaseWireFormat

LOGGER = logging.getLogger(__name__)


class V2PackWireFormat(BaseWireFormat):
    """DIDComm V2 message parser and serializer."""

    async def parse_message(
        self,
        session: ProfileSession,
        message_body: Union[str, bytes],
    ) -> Tuple[dict, MessageReceipt]:
        """Parse message."""

        messaging = session.inject(DIDCommMessaging)

        receipt = MessageReceipt()
        receipt.in_time = time_now()
        receipt.raw_message = message_body
        receipt.didcomm_version = DIDCommVersion.v2

        message_dict = None
        message_json = message_body

        if not message_json:
            raise WireFormatParseError("Message body is empty")

        try:
            message_dict = json.loads(message_json)
        except ValueError:
            raise WireFormatParseError("Message JSON parsing failed")
        if not isinstance(message_dict, dict):
            raise WireFormatParseError("Message JSON result is not an object")

        wallet = session.inject_or(BaseWallet)
        if not wallet:
            LOGGER.error("No wallet found")
            raise WalletNotFoundError()

        # packed messages are detected by the absence of type
        if "type" not in message_dict:
            try:
                message_unpack = await messaging.unpack(message_json)
            except CryptoServiceError:
                LOGGER.info("Message unpack failed, falling back to JSON")
            else:
                # Set message_dict to be the dictionary that we unpacked
                message_dict = message_unpack.message

                if message_unpack.sender_kid:
                    receipt.sender_verkey = message_unpack.sender_kid
                    receipt.recipient_verkey = message_unpack.recipient_kid

        thid = message_dict.get("thid")
        receipt.thread_id = thid or message_dict.get("id")

        receipt.parent_thread_id = message_dict.get("pthid")

        # handle transport decorator
        receipt.direct_response_mode = message_dict.get("return_route")

        LOGGER.debug("Expanded message: %s", message_dict)

        return message_dict, receipt

    async def encode_message(
        self,
        session: ProfileSession,
        message_json: Union[str, bytes],
        recipient_keys: Sequence[str],
        routing_keys: Sequence[str],
        sender_key: str,
    ) -> Union[str, bytes]:
        """Encode an outgoing message for transport.

        Args:
            session: The profile session for providing wallet access
            message_json: The message body to serialize
            recipient_keys: A sequence of recipient verkeys
            routing_keys: A sequence of routing verkeys
            sender_key: The verification key of the sending agent

        Returns:
            The encoded message

        Raises:
            MessageEncodeError: If the message could not be encoded

        """
        messaging = session.inject(DIDCommMessaging)

        if sender_key and recipient_keys:
            message = await messaging.pack(
                message=message_json,
                to=recipient_keys[0].split("#")[0],
                frm=sender_key.split("#")[0],
            )
            message = message.message
        else:
            message = message_json
        return message



"""test prove.py"""

import pytest
import pytest_asyncio
from anoncreds import CredentialRevocationState, RevocationStatusList

from ....anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ....anoncreds.registry import AnonCredsRegistry
from ....anoncreds.tests.mock_objects import MOCK_W3CPRES
from ....resolver.default.key import KeyDIDResolver
from ....resolver.did_resolver import DIDResolver
from ....revocation.models.revocation_registry import RevocationRegistry
from ....tests import mock
from ....utils.testing import create_test_profile
from ....vc.ld_proofs.error import LinkedDataProofException
from ....wallet.default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    DefaultVerificationKeyStrategy,
)
from ....wallet.did_method import DIDMethods
from ...ld_proofs.document_loader import DocumentLoader
from ..prove import (
    _extract_cred_idx,
    _get_predicate_type_and_value,
    _load_w3c_credentials,
    create_rev_states,
    create_signed_anoncreds_presentation,
)
from .test_manager import VC


@pytest.fixture
def resolver():
    yield DIDResolver([KeyDIDResolver()])


@pytest_asyncio.fixture
async def profile(resolver: DIDResolver):
    profile = await create_test_profile()
    profile.context.injector.bind_instance(DIDMethods, DIDMethods())
    profile.context.injector.bind_instance(
        BaseVerificationKeyStrategy, DefaultVerificationKeyStrategy()
    )
    profile.context.injector.bind_instance(DIDResolver, resolver)
    profile.context.injector.bind_instance(DocumentLoader, DocumentLoader(profile))
    yield profile


@pytest.mark.asyncio
async def test_create_signed_anoncreds_presentation():
    profile = await create_test_profile({"wallet.type": "askar-anoncreds"})
    profile.context.injector.bind_instance(DIDMethods, DIDMethods())
    profile.context.injector.bind_instance(
        BaseVerificationKeyStrategy, DefaultVerificationKeyStrategy()
    )
    profile.context.injector.bind_instance(DIDResolver, DIDResolver([KeyDIDResolver()]))
    profile.context.injector.bind_instance(DocumentLoader, DocumentLoader(profile))
    registry = mock.MagicMock(AnonCredsRegistry, autospec=True)
    registry.get_schema = mock.CoroutineMock(
        return_value=mock.MagicMock(
            schema=mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "issuerId": "TNuyNH2pAW2G6z3BW8ZYLf",
                        "attrNames": [
                            "degree",
                            "name",
                            "date",
                            "birthdate_dateint",
                            "timestamp",
                        ],
                        "name": "degree schema",
                        "version": "68.37.38",
                    }
                )
            )
        ),
    )
    registry.get_credential_definition = mock.CoroutineMock(
        return_value=mock.MagicMock(
            credential_definition=mock.MagicMock(
                serialize=mock.MagicMock(
                    return_value={
                        "issuerId": "TNuyNH2pAW2G6z3BW8ZYLf",
                        "schemaId": "1242274",
                        "type": "CL",
                        "tag": "faber.agent.degree_schema",
                        "value": {
                            "primary": {
                                "n": "110191895107383177944225186787127045472400456419044508847920073749118325816882879555888693590503408107531218246551508200778858813446764597252042460295740219285296989541107769089502181778576777417930436325553990675031577262316859960188786027427415918907113489291521324039356114616863101480866513302357191568489145237650236954578848085540776058357840327446186761324850043360872510240168860114128657413033422322367730714244593033343815255367880227269514752281182456165390304117532268729405376542376337985308310279351286237047011259755660096523481882637528390140070572293678914392133356847907443964699396666368223294158061",
                                "s": "9415812539608195450966883098870652296916491635505203853080983952632805384302967432306788150605939803147875077062094749099736966082292346688785945337831356656890556951983186155154365222554857316130536164238857088110797546973073757951768643832909466949503368156244645826883930998004092294090171380523954224832439522626754713977466892117127635439861663437421962242296372559287964629293301712321588335521469805244310171886131183734977508209622245349508899886419257943124600270414702700150921447687226557458680951807835253145114264110732291146813141269571018474399690170085303553149912953395770110046436238930438812780147",
                                "r": {
                                    "birthdate_dateint": "39968681640304788380009477273748351837529664198759676077165379337792646675901616432152370797900826515094147804138445249213613945496792933661400605498715980120831712924000357777147499418810050923414071460841739245157420435453468918475700941287524373324079549579873246517784186882218530939272891114746579591872034727864365855842228114509998007945460010480245507010064392110362136591952150320820422225826544285688677068642300714078641169340591902569248592823824054227933825730852985690007682552923384746004484874059862968040467406436797474876079422807683482900156405231629390843729772887675489043282470807407031205186969",
                                    "date": "1189483679228280739072650972354053068606455987118485184085207434597201188158334235777355692068087062063721870127753534657818335769737924323208415920434077874887496282538032440486476302782516445945959644735720202525994434469144049628428271543868582987595621454107048590112351992445017094207270819619996701213546548131486236291413202271720647291688756231888188593794699964380031588638888843772851461391755221206371940247603348559507718057920345380550594583486893575953718560696958190724499821774481967602504818394556262808918356126374081669920344874855870303258253201308390418654588788321421245107716899898452672677669",
                                    "degree": "7930507600973622255761968698285828306846788291380201483787889093576185243433070380250482080700646211095178134232253270069460369492303935894012408118127081015494559256719242506733590055152439637360319677447051969350390638377492011800932771226273035358153040455317319159092480932337395314046192165987582330687751669998427678060271101315462477794016231639142468258481489055468781240927905148904953855558036223028111290216890655943559980456204339868018478633494472724669957163721224817260017913056937034080193667933454564153001308165996679854735603451224195324968916426779531862123459043377650779177644556309703126058569",
                                    "master_secret": "276481248076029901397915904197571939766835573218052607192738620704710670905748562925445909482694999303080385244124333488797597869947177554955868568791587429441803954589662661470453559603193697529455695636193068394109938416082954992003652270737665467573503772195889460315616035410445104017129549757818399000451038212677658874397388131875030264607848961441095008175483303770912033940379028110558590125358672540783519319152844457014577027814244423449014637405165845051430105495276716743166413876676529214061455612429316323214502293979149015655581759299623912651153921930483011898615547608936905586870150072549489722497",
                                    "name": "82868656041023829484169603686526910742809130312561664067502441442293881789487387499604367473067725155834966664040618821809499511114213661657225202206321891616189181748382445256076664775972124872336589185892368887947537027000430124719570575062210504764164198906984696516109885071145245277063304337918433238788453702459065343194907264612324316538889460640366702448456177503504044952464912882387589287029922982523474901886018581567364020613890822736197654463892632953961461903659266302672901927472391491520220790086494421629530104134487848960992291400448418880719301301643211538480715261724678645854026893882662935786454",
                                    "timestamp": "7031267523384255931724229734709668852236981666398664500628609312084184839422714066646111517571504367881906999946906142205880431579116282139457459430490628368074988968893818537384306625760198775803846638688715282693724689063695143781851672316171826804538178049496230705683899355140338093894078404749873276975725352612375324676593660719462779985715395422238940426215008469041353116398836478202492841783866649731302994305254257899941902815383605743664381990123198126148381917220067404201702094151978496034163563512821780561363033330132250212505439974465744473015256287665802141980858496388564369684473390614913213492963",
                                },
                                "rctxt": "57595457065224964384532723809214070177221283698619189163273994795484340422910421565194502650047793930986106754378617657953634148479003200343596849944404829819454759272470409952906484645890312928698071525795674306099322535554040769358088959541527217412947857624113771388855906636033132647358149811523979540215755447957501980890442007680654813076936017670559128454293737368578572852087014890823425797572658699153814350307873126659084777516770578065919757432552875676249519945052600028967301682399735687262474002867432540942797659023173846671235714105519229325988649284822140546012838121486756140885474564667338033508925",
                                "z": "92392248633579159734489059262157658066602172492655376458258465743015720279931884516580613066265472490833888951133646441506435308647071645829490880133890938400531213056849488446865347760761739826520600426315831962473284037732716361460703303047650224879262380737182724193613502180709449361207809530289742931388401783902150303407116160750968583373664158158828048355594343560019377499225431222258609781090866038688202681876195069528239050879910722452167883110680950451692711483330249956590291961929310759534360045884090140017202751268354570741889324005873479145636960045391944928193003593085006050343477314073342608326752",
                            }
                        },
                    }
                )
            )
        )
    )
    profile.context.injector.bind_instance(AnonCredsRegistry, registry)

    with mock.patch.object(
        AnonCredsHolder, "create_presentation_w3c", return_value=MOCK_W3CPRES
    ) as mock_create_pres:
        await create_signed_anoncreds_presentation(
            profile=profile,
            pres_definition={
                "id": "5591656f-5b5d-40f8-ab5c-9041c8e3a6a0",
                "name": "Age Verification",
                "purpose": "We need to verify your age before entering a bar",
                "format": {
                    "di_vc": {
                        "proof_type": ["DataIntegrityProof"],
                        "cryptosuite": ["anoncreds-2023", "eddsa-rdfc-2022"],
                    }
                },
                "input_descriptors": [
                    {
                        "id": "age-verification",
                        "name": "A specific type of VC + Issuer",
                        "purpose": "We want a VC of this type generated by this issuer",
                        "constraints": {
                            "limit_disclosure": "required",
                            "fields": [
                                {
                                    "path": ["$.issuer"],
                                    "filter": {
                                        "type": "string",
                                        "const": "7yDP6qARVAp1Rims8Fj43k",
                                    },
                                },
                                {"path": ["$.credentialSubject.name"]},
                                {"path": ["$.credentialSubject.degree"]},
                                {
                                    "path": ["$.credentialSubject.birthdate_dateint"],
                                    "predicate": "preferred",
                                    "filter": {"type": "number", "maximum": 20060711},
                                },
                            ],
                            "statuses": {
                                "active": {"directive": "disallowed"},
                                "suspended": {"directive": None},
                                "revoked": {"directive": None},
                            },
                        },
                        "schema": {
                            "uri_groups": [
                                [
                                    {
                                        "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                                    }
                                ]
                            ],
                            "oneof_filter": False,
                        },
                    }
                ],
            },
            presentation={
                "@context": ["https://www.w3.org/2018/credentials/v1"],
                "type": ["VerifiablePresentation"],
                "verifiableCredential": [
                    {
                        "@context": [
                            "https://www.w3.org/2018/credentials/v1",
                            "https://w3id.org/security/data-integrity/v2",
                            {
                                "@vocab": "https://www.w3.org/ns/credentials/issuer-dependent#"
                            },
                        ],
                        "type": ["VerifiableCredential"],
                        "issuer": "7yDP6qARVAp1Rims8Fj43k",
                        "issuanceDate": "2024-07-11T18:33:24.564345117Z",
                        "credentialSubject": {
                            "name": "Alice Smith",
                            "date": "2018-05-28",
                            "timestamp": 1720722800,
                            "degree": "Maths",
                            "birthdate_dateint": 20000711,
                        },
                        "proof": {
                            "type": "DataIntegrityProof",
                            "proofPurpose": "assertionMethod",
                            "verificationMethod": "7yDP6qARVAp1Rims8Fj43k:3:CL:1255455:faber.agent.degree_schema",
                            "proofValue": "ukgGEqXNjaGVtYV9pZNkvN3lEUDZxQVJWQXAxUmltczhGajQzazoyOmRlZ3JlZSBzY2hlbWE6NDYuMzkuNzmrY3JlZF9kZWZfaWTZPTd5RFA2cUFSVkFwMVJpbXM4Rmo0M2s6MzpDTDoxMjU1NDU1OmZhYmVyLmFnZW50LmRlZ3JlZV9zY2hlbWGpc2lnbmF0dXJlgqxwX2NyZWRlbnRpYWyEo21fMtwAIFHM9cyPzOMsEls6E3zMtC0BaMzTGsyazNzMnsz2zL9lzJHMj8zHaczATsyUYi7Mv6Fh3AEBAsy8zMrMwFHM5MzHzPTMxQc-zPjM68ywV8ypXszEzJLMv8yaNczYDhvMqR7M9CzMv8yweszIccyqCWHMiwbMrGrMyRU0zOTM2sypBczfzII7VksbBMzXTRzMyhjM2idFNsyCBszDOMzVzLDM3cy8IVDMy0YETSfM_VFwzPfM-gdaAcyUNBJ8zPxgeszLO8zwzNjM98zPNFgKBRXM_mZ4zIXMnszOzOjMscyNJQTMmszjD3drzMXMgSBAzJ_M8MzkawYXzP7MrszKzLJPzPXMjEd8HsyzEczBzNEmzJ7Mkl_Mmcz0csziGk9eVD9ezMEfchjMzsyFJsyhQj_MrczyzLDMtWYczIorzMULzPbM0mUSzP1aR8zvzJQRQcyuzP1czK8TDQxSK8zjzOZ8zKIFUcyXzPZpdk9PzLpLNczzHyTM7SU5zJQFRcyaWAdcWczodSfMkszKAmLM7igjeMz_zN3Ms8zAT0TM2cyYJmEqzODM20rM8klTzPChZdwASxAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMozMsWzN3M8MzSDg3MhsyTzLRLzIwzoXbcAVUMzM3MlCDMhMyMH1LMigd7JEUHIMzuW8yQJT8iAm7Mpw5nAsyTzKbMtsyUzPUzzK9FzLLMkFLMzynMu8yEzN5nO8z3FjvMiszIzJbM3cywWMzTEszQzPzMx8yNBcycexcle8yIzJfMycyTQszUMBTMg8zGzNPMisz3zLhnEszDdcyozLXM1TYFzPRCO8zfcSQLNGPMoszyf8yfCwtVzNbM18zJzNDMscyzzJbM3l3Mh8zPzKYhCcyXzIrM98ykzODMiczSzJPMgRXMgnDMhkLMhCLMkwRkzKotF8yDzPbMmGzMoC4LRMyZzJ3M8w7MvczHLcyzzMTMmCdJzILMwUHM98zMzPpSDyLM5X7M2kXMg8yQXHvM8MyqV0nMmAgQzKx9LTM2V8zlzLLMv8z8zPJqTMzjR8zLFlU2KFRHc8zXzOfMrMzSdArM7CDMksykzM3MjMznLGo-zLzM7QVqzPXM6Bl9aMzgzIsSzLlpNMyYa8z5zId1BnNyzPUxZczyX8zSzNYKfMzIzJPM2QvM_MzIzLvMvRw4ZgwgzLkvGszAKszmMMzfzLbM4czQRGfM21Z5PzsxzIRlzJETzNszMMydzOdAzObMuczJzJUZzKgXzJ_MlG0PDcz4QCXMpwDMyszXzPMVzOFrzKjM3hc6zP_M-sz3XMyfS1vMvF_Mlsz7MMzozKzMk8y0rHJfY3JlZGVudGlhbMC7c2lnbmF0dXJlX2NvcnJlY3RuZXNzX3Byb29mgqJzZdwBAMy0zO7MyszkzN7MxcyVQMy3zNEdzKBvzMfMq2Q5agJwG8yazPjMpMyQzJ1JV8z_zOrMkMzWzKPMjcybPMyHX8yVzMDM-WBwBcz6zN0yHRbMnSDMoMyXHMynzOfMlMyiCnfM-Mylb8zKzPBHzMrMm1TM2l7Mp8zCQMyyzIYHzMBXzPJFf04lzPpYzKLM7sykKMzHzJgJzNoTJEVtzP5KzJ7Mz8yAXUPM-13M18yQWhpOXszRFczqzI3MqszDJ8yjE2V-zLxUzPpDUknMrh_Mu0sVzJBAzMgqfX1ZF8zRzLzMr8zIzKtdeRbM4szqCFnMqczXdsyIPMzXzK_MhQTM5sypzOrM5MziHlXMxxrMpMzTzPjMjXbMzEZxNsykzJRKFjXM-XtYzIcEzJzMgMynzITMyWxpEix9zOvM1cy5ZcyacMzJLMzxzN9OzM3MuV7M28z7zKQRHsyXzPDMvczhB8zgzNrMmxwhJMydEhPMhUkvzIHM3cytzL8CIiLMtczJzOLMlMyZOMzOzLcizOwAoWPcACAvzI_MrkTMvDvMmMy2zIvMizDM0ihSzMd_VGDMvMzEWn3MrczcbMykzKNqzLgzzKvMzw",
                            "cryptosuite": "anoncreds-2023",
                        },
                    }
                ],
                "presentation_submission": {
                    "id": "700e4ed4-bb73-403f-a73c-e456556cbdf0",
                    "definition_id": "5591656f-5b5d-40f8-ab5c-9041c8e3a6a0",
                    "descriptor_map": [
                        {
                            "id": "age-verification",
                            "format": "ldp_vc",
                            "path": "$.verifiableCredential[0]",
                        }
                    ],
                },
            },
            challenge="3fa85f64-5717-4562-b3fc-2c963f66afa7",
            domain="domain",
            credentials=[
                {
                    "@context": [
                        "https://www.w3.org/2018/credentials/v1",
                        "https://w3id.org/security/data-integrity/v2",
                        {"@vocab": "https://www.w3.org/ns/credentials/issuer-dependent#"},
                    ],
                    "type": ["VerifiableCredential"],
                    "issuer": "7yDP6qARVAp1Rims8Fj43k",
                    "issuanceDate": "2024-07-11T18:33:24.564345117Z",
                    "credentialSubject": {
                        "name": "Alice Smith",
                        "date": "2018-05-28",
                        "timestamp": 1720722800,
                        "degree": "Maths",
                        "birthdate_dateint": 20000711,
                    },
                    "proof": {
                        "type": "DataIntegrityProof",
                        "proofPurpose": "assertionMethod",
                        "verificationMethod": "7yDP6qARVAp1Rims8Fj43k:3:CL:1255455:faber.agent.degree_schema",
                        "proofValue": "ukgGEqXNjaGVtYV9pZNkvN3lEUDZxQVJWQXAxUmltczhGajQzazoyOmRlZ3JlZSBzY2hlbWE6NDYuMzkuNzmrY3JlZF9kZWZfaWTZPTd5RFA2cUFSVkFwMVJpbXM4Rmo0M2s6MzpDTDoxMjU1NDU1OmZhYmVyLmFnZW50LmRlZ3JlZV9zY2hlbWGpc2lnbmF0dXJlgqxwX2NyZWRlbnRpYWyEo21fMtwAIFHM9cyPzOMsEls6E3zMtC0BaMzTGsyazNzMnsz2zL9lzJHMj8zHaczATsyUYi7Mv6Fh3AEBAsy8zMrMwFHM5MzHzPTMxQc-zPjM68ywV8ypXszEzJLMv8yaNczYDhvMqR7M9CzMv8yweszIccyqCWHMiwbMrGrMyRU0zOTM2sypBczfzII7VksbBMzXTRzMyhjM2idFNsyCBszDOMzVzLDM3cy8IVDMy0YETSfM_VFwzPfM-gdaAcyUNBJ8zPxgeszLO8zwzNjM98zPNFgKBRXM_mZ4zIXMnszOzOjMscyNJQTMmszjD3drzMXMgSBAzJ_M8MzkawYXzP7MrszKzLJPzPXMjEd8HsyzEczBzNEmzJ7Mkl_Mmcz0csziGk9eVD9ezMEfchjMzsyFJsyhQj_MrczyzLDMtWYczIorzMULzPbM0mUSzP1aR8zvzJQRQcyuzP1czK8TDQxSK8zjzOZ8zKIFUcyXzPZpdk9PzLpLNczzHyTM7SU5zJQFRcyaWAdcWczodSfMkszKAmLM7igjeMz_zN3Ms8zAT0TM2cyYJmEqzODM20rM8klTzPChZdwASxAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMozMsWzN3M8MzSDg3MhsyTzLRLzIwzoXbcAVUMzM3MlCDMhMyMH1LMigd7JEUHIMzuW8yQJT8iAm7Mpw5nAsyTzKbMtsyUzPUzzK9FzLLMkFLMzynMu8yEzN5nO8z3FjvMiszIzJbM3cywWMzTEszQzPzMx8yNBcycexcle8yIzJfMycyTQszUMBTMg8zGzNPMisz3zLhnEszDdcyozLXM1TYFzPRCO8zfcSQLNGPMoszyf8yfCwtVzNbM18zJzNDMscyzzJbM3l3Mh8zPzKYhCcyXzIrM98ykzODMiczSzJPMgRXMgnDMhkLMhCLMkwRkzKotF8yDzPbMmGzMoC4LRMyZzJ3M8w7MvczHLcyzzMTMmCdJzILMwUHM98zMzPpSDyLM5X7M2kXMg8yQXHvM8MyqV0nMmAgQzKx9LTM2V8zlzLLMv8z8zPJqTMzjR8zLFlU2KFRHc8zXzOfMrMzSdArM7CDMksykzM3MjMznLGo-zLzM7QVqzPXM6Bl9aMzgzIsSzLlpNMyYa8z5zId1BnNyzPUxZczyX8zSzNYKfMzIzJPM2QvM_MzIzLvMvRw4ZgwgzLkvGszAKszmMMzfzLbM4czQRGfM21Z5PzsxzIRlzJETzNszMMydzOdAzObMuczJzJUZzKgXzJ_MlG0PDcz4QCXMpwDMyszXzPMVzOFrzKjM3hc6zP_M-sz3XMyfS1vMvF_Mlsz7MMzozKzMk8y0rHJfY3JlZGVudGlhbMC7c2lnbmF0dXJlX2NvcnJlY3RuZXNzX3Byb29mgqJzZdwBAMy0zO7MyszkzN7MxcyVQMy3zNEdzKBvzMfMq2Q5agJwG8yazPjMpMyQzJ1JV8z_zOrMkMzWzKPMjcybPMyHX8yVzMDM-WBwBcz6zN0yHRbMnSDMoMyXHMynzOfMlMyiCnfM-Mylb8zKzPBHzMrMm1TM2l7Mp8zCQMyyzIYHzMBXzPJFf04lzPpYzKLM7sykKMzHzJgJzNoTJEVtzP5KzJ7Mz8yAXUPM-13M18yQWhpOXszRFczqzI3MqszDJ8yjE2V-zLxUzPpDUknMrh_Mu0sVzJBAzMgqfX1ZF8zRzLzMr8zIzKtdeRbM4szqCFnMqczXdsyIPMzXzK_MhQTM5sypzOrM5MziHlXMxxrMpMzTzPjMjXbMzEZxNsykzJRKFjXM-XtYzIcEzJzMgMynzITMyWxpEix9zOvM1cy5ZcyacMzJLMzxzN9OzM3MuV7M28z7zKQRHsyXzPDMvczhB8zgzNrMmxwhJMydEhPMhUkvzIHM3cytzL8CIiLMtczJzOLMlMyZOMzOzLcizOwAoWPcACAvzI_MrkTMvDvMmMy2zIvMizDM0ihSzMd_VGDMvMzEWn3MrczcbMykzKNqzLgzzKvMzw",
                        "cryptosuite": "anoncreds-2023",
                    },
                }
            ],
            purpose="assertionMethod",
        )

        mock_create_pres.assert_called_once()


def test__extract_cred_idx():
    item_path = "$.verifiableCredential[0]"
    assert _extract_cred_idx(item_path) == 0

    item_path = "$.verifiableCredential[42]"
    assert _extract_cred_idx(item_path) == 42


def test__get_predicate_type_and_value():
    pred_filter: dict[str, int] = {"exclusiveMinimum": 10}
    assert _get_predicate_type_and_value(pred_filter) == (">", 10)

    pred_filter = {"exclusiveMaximum": 20}
    assert _get_predicate_type_and_value(pred_filter) == ("<", 20)

    pred_filter = {"minimum": 5}
    assert _get_predicate_type_and_value(pred_filter) == (">=", 5)

    pred_filter = {"maximum": 15}
    assert _get_predicate_type_and_value(pred_filter) == ("<=", 15)


@pytest.mark.asyncio
async def test__load_w3c_credentials():
    credentials = [VC]

    w3c_creds = await _load_w3c_credentials(credentials)

    assert len(w3c_creds) == len(credentials)

    with pytest.raises(LinkedDataProofException):
        credentials = [{"schema": "invalid"}]
        await _load_w3c_credentials(credentials)


@pytest.mark.asyncio
async def test_create_rev_states():
    w3c_creds_metadata = [
        {"rev_reg_id": "rev_reg_id_1", "rev_reg_index": 0, "timestamp": 1234567890},
        {"rev_reg_id": "rev_reg_id_2", "rev_reg_index": 1, "timestamp": 1234567890},
    ]
    rev_reg_defs = {
        "rev_reg_id_1": {"id": "rev_reg_id_1", "definition": "rev_reg_def_1"},
        "rev_reg_id_2": {"id": "rev_reg_id_2", "definition": "rev_reg_def_2"},
    }
    rev_reg_entries = {
        "rev_reg_id_1": {1234567890: "rev_reg_entry_1"},
        "rev_reg_id_2": {1234567890: "rev_reg_entry_2"},
    }

    with mock.patch.object(
        RevocationRegistry,
        "from_definition",
        return_value=mock.CoroutineMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(return_value="tails_path")
        ),
    ):
        with mock.patch.object(
            RevocationStatusList, "load", return_value=mock.MagicMock()
        ):
            with pytest.raises(AnonCredsHolderError):
                await create_rev_states(w3c_creds_metadata, rev_reg_defs, rev_reg_entries)
            with mock.patch.object(
                CredentialRevocationState, "create", return_value=mock.MagicMock()
            ) as mock_create:
                result = await create_rev_states(
                    w3c_creds_metadata, rev_reg_defs, rev_reg_entries
                )

                assert len(result) == 2
                assert mock_create.call_count == 2



"""Cancel transaction request handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import TransactionManager, TransactionManagerError
from ..messages.cancel_transaction import CancelTransaction


class TransactionCancelHandler(BaseHandler):
    """Handler class for Cancel transaction request."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle cancel transaction request.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug("TransactionCancelHandler called with context %s", context)
        assert isinstance(context.message, CancelTransaction)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        mgr = TransactionManager(context.profile)
        try:
            await mgr.receive_cancel_transaction(
                context.message, context.connection_record.connection_id
            )
        except TransactionManagerError:
            self._logger.exception("Error receiving cancel transaction request")



TEST_LD_DOCUMENT_BAD_SIGNED_ED25519_2020 = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
    "proof": {
        "type": "Ed25519Signature2020",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2023-05-24T13:52:12.985755+00:00",
        "proofPurpose": "assertionMethod",
        "proofValue": "z4JqUKXM2hj49WyCFmh8aUMgsEMfNJBd3WN3xGsowBPC6s3pPdJTgBoKcWueN7amex9wVdWGZNS8p8d38sqr7ccWL",
    },
}



"""Utilities for testing."""

from typing import Optional
from uuid import uuid4

from ..askar.profile import AskarProfile
from ..askar.profile_anon import AskarAnonCredsProfile
from ..askar.store import AskarStoreConfig
from ..config.injection_context import InjectionContext


async def create_test_profile(
    settings: Optional[dict] = None, context: Optional[InjectionContext] = None
) -> AskarProfile | AskarAnonCredsProfile:
    """Create a profile for testing."""
    if not settings:
        settings = {
            "wallet.type": "askar",
            "auto_provision": True,
            "wallet.key": "5BngFuBpS4wjFfVFCtPqoix3ZXG2XR8XJ7qosUzMak7R",
            "wallet.key_derivation_method": "RAW",
        }
    _id = settings.get("wallet.id", str(uuid4()))
    settings["wallet.id"] = _id
    """Create a profile for testing."""
    store_config = AskarStoreConfig(
        {
            "name": _id,
            "key": "5BngFuBpS4wjFfVFCtPqoix3ZXG2XR8XJ7qosUzMak7R",
            "key_derivation_method": "RAW",
            "auto_recreate": True,
        }
    )
    if not context:
        context = InjectionContext(
            settings=settings,
        )
    opened = await store_config.open_store(provision=True, in_memory=True)

    if settings.get("wallet.type") == "askar-anoncreds":
        return AskarAnonCredsProfile(
            opened=opened,
            context=context,
        )
    return AskarProfile(
        opened=opened,
        context=context,
    )



"""Holder admin routes."""

import json
from profile import Profile

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from aries_askar import AskarErrorCode
from marshmallow import fields
from marshmallow.validate import Range

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ..indy.holder import IndyHolder, IndyHolderError
from ..indy.models.cred_precis import IndyCredInfoSchema
from ..ledger.base import BaseLedger
from ..ledger.error import LedgerError
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.valid import (
    ENDPOINT_EXAMPLE,
    ENDPOINT_VALIDATE,
    INDY_WQL_EXAMPLE,
    INDY_WQL_VALIDATE,
    NUM_STR_NATURAL_EXAMPLE,
    NUM_STR_NATURAL_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
    UUID4_EXAMPLE,
)
from ..storage.base import DEFAULT_PAGE_SIZE, MAXIMUM_PAGE_SIZE
from ..storage.error import StorageError, StorageNotFoundError
from ..storage.vc_holder.base import VCHolder
from ..storage.vc_holder.vc_record import VCRecordSchema
from ..wallet.error import WalletNotFoundError

wallet_type_config = "wallet.type"


class HolderModuleResponseSchema(OpenAPISchema):
    """Response schema for Holder Module."""


class AttributeMimeTypesResultSchema(OpenAPISchema):
    """Result schema for credential attribute MIME type."""

    results = fields.Dict(
        keys=fields.Str(metadata={"description": "Attribute name"}),
        values=fields.Str(metadata={"description": "MIME type"}),
        allow_none=True,
    )


class CredInfoListSchema(OpenAPISchema):
    """Result schema for credential query."""

    results = fields.List(fields.Nested(IndyCredInfoSchema()))


class CredentialsListQueryStringSchema(OpenAPISchema):
    """Parameters and validators for query string in credentials list query."""

    start = fields.Str(
        required=False,
        load_default=0,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Start index (DEPRECATED - use offset instead)",
            "example": NUM_STR_WHOLE_EXAMPLE,
            "deprecated": True,
        },
    )
    count = fields.Str(
        required=False,
        load_default=10,
        validate=NUM_STR_NATURAL_VALIDATE,
        metadata={
            "description": "Maximum number to retrieve (DEPRECATED - use limit instead)",
            "example": NUM_STR_NATURAL_EXAMPLE,
            "deprecated": True,
        },
    )
    limit = fields.Int(
        required=False,
        validate=Range(min=1, max=MAXIMUM_PAGE_SIZE),
        metadata={"description": "Number of results to return", "example": 50},
    )
    offset = fields.Int(
        required=False,
        validate=Range(min=0),
        metadata={"description": "Offset for pagination", "example": 0},
    )
    wql = fields.Str(
        required=False,
        validate=INDY_WQL_VALIDATE,
        metadata={"description": "(JSON) WQL query", "example": INDY_WQL_EXAMPLE},
    )


class W3CCredentialsListRequestSchema(OpenAPISchema):
    """Parameters and validators for W3C credentials request."""

    contexts = fields.List(
        fields.Str(
            validate=ENDPOINT_VALIDATE,
            metadata={
                "description": "Credential context to match",
                "example": ENDPOINT_EXAMPLE,
            },
        ),
        required=False,
    )
    types = fields.List(
        fields.Str(
            validate=ENDPOINT_VALIDATE,
            metadata={
                "description": "Credential type to match",
                "example": ENDPOINT_EXAMPLE,
            },
        ),
        required=False,
    )
    schema_ids = fields.List(
        fields.Str(
            validate=ENDPOINT_VALIDATE,
            metadata={
                "description": "Credential schema identifier",
                "example": ENDPOINT_EXAMPLE,
            },
        ),
        required=False,
        metadata={"description": "Schema identifiers, all of which to match"},
    )
    issuer_id = fields.Str(
        required=False,
        metadata={"description": "Credential issuer identifier to match"},
    )
    subject_ids = fields.List(
        fields.Str(metadata={"description": "Subject identifier"}),
        required=False,
        metadata={"description": "Subject identifiers, all of which to match"},
    )
    proof_types = fields.List(
        fields.Str(
            metadata={
                "description": "Signature suite used for proof",
                "example": "Ed25519Signature2018",
            }
        )
    )
    given_id = fields.Str(
        required=False, metadata={"description": "Given credential id to match"}
    )
    tag_query = fields.Dict(
        keys=fields.Str(metadata={"description": "Tag name"}),
        values=fields.Str(metadata={"description": "Tag value"}),
        required=False,
        metadata={"description": "Tag filter"},
    )
    max_results = fields.Int(
        required=False,
        metadata={"strict": True, "description": "Maximum number of results to return"},
    )


class VCRecordListSchema(OpenAPISchema):
    """Result schema for W3C credential query."""

    results = fields.List(fields.Nested(VCRecordSchema()))


class HolderCredIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential id."""

    credential_id = fields.Str(
        required=True,
        metadata={"description": "Credential identifier", "example": UUID4_EXAMPLE},
    )


class CredRevokedQueryStringSchema(OpenAPISchema):
    """Path parameters and validators for request seeking cred revocation status."""

    fro = fields.Str(
        data_key="from",
        required=False,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Earliest epoch of revocation status interval of interest",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )
    to = fields.Str(
        required=False,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Latest epoch of revocation status interval of interest",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )


class CredRevokedResultSchema(OpenAPISchema):
    """Result schema for credential revoked request."""

    revoked = fields.Bool(
        metadata={"description": "Whether credential is revoked on the ledger"}
    )


@docs(tags=["credentials"], summary="Fetch credential from wallet by id")
@match_info_schema(HolderCredIdMatchInfoSchema())
@response_schema(IndyCredInfoSchema(), 200, description="")
@tenant_authentication
async def credentials_get(request: web.BaseRequest):
    """Request handler for retrieving credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential info

    """
    context: AdminRequestContext = request["context"]
    credential_id = request.match_info["credential_id"]

    if context.settings.get(wallet_type_config) == "askar-anoncreds":
        holder = AnonCredsHolder(context.profile)
    else:
        holder = context.profile.inject(IndyHolder)

    try:
        credential = await holder.get_credential(credential_id)
    except WalletNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    credential_json = json.loads(credential)
    return web.json_response(credential_json)


@docs(tags=["credentials"], summary="Query credential revocation status by id")
@match_info_schema(HolderCredIdMatchInfoSchema())
@querystring_schema(CredRevokedQueryStringSchema())
@response_schema(CredRevokedResultSchema(), 200, description="")
@tenant_authentication
async def credentials_revoked(request: web.BaseRequest):
    """Request handler for querying revocation status of credential.

    Args:
        request: aiohttp request object

    Returns:
        Empty production

    """
    context: AdminRequestContext = request["context"]
    credential_id = request.match_info["credential_id"]
    fro = request.query.get("from")
    to = request.query.get("to")
    profile = context.profile
    wallet_type = profile.settings.get_value(wallet_type_config)

    async def get_revoked_using_anoncreds(profile: Profile):
        holder = AnonCredsHolder(profile)
        return await holder.credential_revoked(
            credential_id,
            int(fro) if fro else None,
            int(to) if to else None,
        )

    async def get_revoked_using_indy(profile: Profile):
        async with profile.session() as session:
            ledger = session.inject_or(BaseLedger)
            if not ledger:
                raise web.HTTPForbidden(reason="No ledger available")

            holder = session.inject(IndyHolder)

            async with ledger:
                try:
                    return await holder.credential_revoked(
                        ledger,
                        credential_id,
                        int(fro) if fro else None,
                        int(to) if to else None,
                    )
                except LedgerError as err:
                    raise web.HTTPBadRequest(reason=err.roll_up) from err

    try:
        if wallet_type == "askar-anoncreds":
            revoked = await get_revoked_using_anoncreds(profile)
        else:
            revoked = await get_revoked_using_indy(profile)
    except WalletNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response({"revoked": revoked})


@docs(tags=["credentials"], summary="Get attribute MIME types from wallet")
@match_info_schema(HolderCredIdMatchInfoSchema())
@response_schema(AttributeMimeTypesResultSchema(), 200, description="")
@tenant_authentication
async def credentials_attr_mime_types_get(request: web.BaseRequest):
    """Request handler for getting credential attribute MIME types.

    Args:
        request: aiohttp request object

    Returns:
        The MIME types response

    """
    context: AdminRequestContext = request["context"]
    credential_id = request.match_info["credential_id"]

    if context.settings.get(wallet_type_config) == "askar-anoncreds":
        holder = AnonCredsHolder(context.profile)
        mime_types = await holder.get_mime_type(credential_id)
    else:
        async with context.profile.session() as session:
            holder = session.inject(IndyHolder)
            mime_types = await holder.get_mime_type(credential_id)

    return web.json_response({"results": mime_types})


@docs(tags=["credentials"], summary="Remove credential from wallet by id")
@match_info_schema(HolderCredIdMatchInfoSchema())
@response_schema(HolderModuleResponseSchema(), description="")
@tenant_authentication
async def credentials_remove(request: web.BaseRequest):
    """Request handler for searching connection records.

    Args:
        request: aiohttp request object

    Returns:
        Empty production

    """
    context: AdminRequestContext = request["context"]
    credential_id = request.match_info["credential_id"]
    profile: Profile = context.profile

    async def delete_credential_using_anoncreds():
        try:
            holder = AnonCredsHolder(profile)
            await holder.delete_credential(credential_id)
        except AnonCredsHolderError as err:
            if err.error_code == AskarErrorCode.NOT_FOUND:
                raise web.HTTPNotFound(reason=err.roll_up) from err
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    async def delete_credential_using_indy():
        async with profile.session() as session:
            try:
                holder = session.inject(IndyHolder)
                await holder.delete_credential(credential_id)
            except WalletNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

    async def delete_using_anoncreds_or_indy():
        """Try to delete anoncreds credential with fallback to indy if not found."""
        try:
            await delete_credential_using_anoncreds()
        except web.HTTPNotFound as anoncreds_err:
            # If credential not found in anoncreds, try with indy
            try:
                await delete_credential_using_indy()
            except web.HTTPNotFound:
                # Raise original anoncreds error if neither found
                raise web.HTTPNotFound(reason=anoncreds_err.reason) from anoncreds_err

    if context.settings.get(wallet_type_config) == "askar-anoncreds":
        await delete_using_anoncreds_or_indy()
    else:
        await delete_credential_using_indy()

    # Notify event subscribers
    topic = "acapy::record::credential::delete"
    await profile.notify(topic, {"id": credential_id, "state": "deleted"})

    return web.json_response({})


@docs(
    tags=["credentials"],
    summary="Fetch credentials from wallet",
)
@querystring_schema(CredentialsListQueryStringSchema())
@response_schema(CredInfoListSchema(), 200, description="")
@tenant_authentication
async def credentials_list(request: web.BaseRequest):
    """Request handler for searching credential records.

    Args:
        request: aiohttp request object

    Returns:
        The credential info list response

    """
    context: AdminRequestContext = request["context"]

    # Handle both old style start/count and new limit/offset
    # TODO: Remove start/count and swap to PaginatedQuerySchema and get_limit_offset
    if "limit" in request.query or "offset" in request.query:
        # New style - use limit/offset
        limit = int(request.query.get("limit", DEFAULT_PAGE_SIZE))
        offset = int(request.query.get("offset", 0))
    else:
        # Old style - use start/count
        limit = int(request.query.get("count", "10"))
        offset = int(request.query.get("start", "0"))

    # url encoded json wql
    encoded_wql = request.query.get("wql") or "{}"
    wql = json.loads(encoded_wql)

    if context.settings.get(wallet_type_config) == "askar-anoncreds":
        holder = AnonCredsHolder(context.profile)
        credentials = await holder.get_credentials(limit=limit, offset=offset, wql=wql)
    else:
        async with context.profile.session() as session:
            holder = session.inject(IndyHolder)
            try:
                credentials = await holder.get_credentials(
                    limit=limit, offset=offset, wql=wql
                )
            except IndyHolderError as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": credentials})


@docs(
    tags=["credentials"],
    summary="Fetch W3C credential from wallet by id",
)
@match_info_schema(HolderCredIdMatchInfoSchema())
@response_schema(VCRecordSchema(), 200, description="")
@tenant_authentication
async def w3c_cred_get(request: web.BaseRequest):
    """Request handler for retrieving W3C credential.

    Args:
        request: aiohttp request object

    Returns:
        Verifiable credential record

    """
    context: AdminRequestContext = request["context"]
    credential_id = request.match_info["credential_id"]

    async with context.profile.session() as session:
        holder = session.inject(VCHolder)
        try:
            vc_record = await holder.retrieve_credential_by_id(credential_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(vc_record.serialize())


@docs(
    tags=["credentials"],
    summary="Remove W3C credential from wallet by id",
)
@match_info_schema(HolderCredIdMatchInfoSchema())
@response_schema(HolderModuleResponseSchema(), 200, description="")
@tenant_authentication
async def w3c_cred_remove(request: web.BaseRequest):
    """Request handler for deleting W3C credential.

    Args:
        request: aiohttp request object

    Returns:
        Empty production

    """
    context: AdminRequestContext = request["context"]
    credential_id = request.match_info["credential_id"]

    async with context.profile.session() as session:
        holder = session.inject(VCHolder)
        try:
            vc_record = await holder.retrieve_credential_by_id(credential_id)
            await holder.delete_credential(vc_record)
            topic = "acapy::record::w3c_credential::delete"
            await session.profile.notify(topic, {"id": credential_id, "state": "deleted"})
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


@docs(
    tags=["credentials"],
    summary="Fetch W3C credentials from wallet",
)
@request_schema(W3CCredentialsListRequestSchema())
@response_schema(VCRecordListSchema(), 200, description="")
@tenant_authentication
async def w3c_creds_list(request: web.BaseRequest):
    """Request handler for searching W3C credential records.

    Args:
        request: aiohttp request object

    Returns:
        The credential record list response

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()
    contexts = body.get("contexts")
    types = body.get("types")
    schema_ids = body.get("schema_ids")
    issuer_id = body.get("issuer_id")
    subject_ids = body.get("subject_ids")
    proof_types = body.get("proof_types")
    given_id = body.get("given_id")
    tag_query = body.get("tag_query")
    max_results = body.get("max_results")

    async with context.profile.session() as session:
        holder = session.inject(VCHolder)
        try:
            search = holder.search_credentials(
                contexts=contexts,
                types=types,
                schema_ids=schema_ids,
                issuer_id=issuer_id,
                subject_ids=subject_ids,
                proof_types=proof_types,
                given_id=given_id,
                tag_query=tag_query,
            )
            records = await search.fetch(max_results)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": [record.serialize() for record in records]})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/credential/{credential_id}", credentials_get, allow_head=False),
            web.get(
                "/credential/revoked/{credential_id}",
                credentials_revoked,
                allow_head=False,
            ),
            web.get(
                "/credential/mime-types/{credential_id}",
                credentials_attr_mime_types_get,
                allow_head=False,
            ),
            web.delete("/credential/{credential_id}", credentials_remove),
            web.get("/credentials", credentials_list, allow_head=False),
            web.get(
                "/credential/w3c/{credential_id}",
                w3c_cred_get,
                allow_head=False,
            ),
            web.delete("/credential/w3c/{credential_id}", w3c_cred_remove),
            web.post("/credentials/w3c", w3c_creds_list),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "credentials",
            "description": "Holder credential management",
            "externalDocs": {
                "description": "Overview",
                "url": "https://w3c.github.io/vc-data-model/#credentials",
            },
        }
    )



"""Classes to manage credential revocation."""

import logging
from typing import Mapping, Optional, Sequence, Text, Tuple

from ..anoncreds.default.legacy_indy.registry import LegacyIndyRegistry
from ..anoncreds.revocation import AnonCredsRevocation
from ..core.error import BaseError
from ..core.profile import Profile
from ..protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange,
)
from ..protocols.issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from ..protocols.revocation_notification.v1_0.models.rev_notification_record import (
    RevNotificationRecord,
)
from ..revocation.util import notify_pending_cleared_event
from ..storage.error import StorageNotFoundError
from .models.issuer_cred_rev_record import IssuerCredRevRecord


class RevocationManagerError(BaseError):
    """Revocation manager error."""


class RevocationManager:
    """Class for managing revocation operations."""

    def __init__(self, profile: Profile):
        """Initialize a RevocationManager.

        Args:
            profile: The profile instance for this revocation manager
        """
        self._profile = profile
        self._logger = logging.getLogger(__name__)

    async def revoke_credential_by_cred_ex_id(
        self,
        cred_ex_id: str,
        publish: bool = False,
        notify: bool = False,
        notify_version: Optional[str] = None,
        thread_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        comment: Optional[str] = None,
        options: Optional[dict] = None,
    ):
        """Revoke a credential by its credential exchange identifier at issue.

        Optionally, publish the corresponding revocation registry delta to the ledger.

        Args:
            cred_ex_id (str): The credential exchange identifier.
            publish (bool, optional): Whether to publish the resulting revocation
                registry delta, along with any revocations pending against it.
                Defaults to False.
            notify (bool, optional): Whether to notify the issuer of the revocation.
                Defaults to False.
            notify_version (str, optional): The version of the revocation notification to
                use. Defaults to None.
            thread_id (str, optional): The thread identifier for the revocation.
                Defaults to None.
            connection_id (str, optional): The connection identifier for the revocation.
                Defaults to None.
            comment (str, optional): A comment for the revocation. Defaults to None.
            options (dict, optional): Additional options for the revocation. Defaults to
                None.

        Raises:
            RevocationManagerError: If no issuer credential revocation record is found
                for the given credential exchange id.

        Returns:
            dict: The result of the credential revocation operation.

        """
        try:
            async with self._profile.session() as session:
                rec = await IssuerCredRevRecord.retrieve_by_cred_ex_id(
                    session,
                    cred_ex_id,
                )
        except StorageNotFoundError as err:
            raise RevocationManagerError(
                "No issuer credential revocation record found for "
                f"credential exchange id {cred_ex_id}"
            ) from err

        return await self.revoke_credential(
            rev_reg_id=rec.rev_reg_id,
            cred_rev_id=rec.cred_rev_id,
            publish=publish,
            notify=notify,
            notify_version=notify_version,
            thread_id=thread_id,
            connection_id=connection_id,
            comment=comment,
            options=options,
        )

    async def revoke_credential(
        self,
        rev_reg_id: str,
        cred_rev_id: str,
        publish: bool = False,
        notify: bool = False,
        notify_version: Optional[str] = None,
        thread_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        comment: Optional[str] = None,
        options: Optional[dict] = None,
    ):
        """Revoke a credential.

        Optionally, publish the corresponding revocation registry delta to the ledger.

        Args:
            rev_reg_id (str): The revocation registry id.
            cred_rev_id (str): The credential revocation id.
            publish (bool, optional): Whether to publish the resulting revocation
                registry delta, along with any revocations pending against it.
                Defaults to False.
            notify (bool, optional): Whether to send a revocation notification.
                Defaults to False.
            notify_version (str, optional): The version of the revocation notification.
                Defaults to None.
            thread_id (str, optional): The thread id for the revocation notification.
                Defaults to None.
            connection_id (str, optional): The connection id for the revocation
                notification. Defaults to None.
            comment (str, optional): A comment for the revocation notification.
                Defaults to None.
            options (dict, optional): Additional options for revocation. Defaults to None.

        Raises:
            RevocationManagerError: If no revocation registry record is found for the
                given rev_reg_id.

        """
        revoc = AnonCredsRevocation(self._profile)
        rev_reg_def = await revoc.get_created_revocation_registry_definition(rev_reg_id)
        if not rev_reg_def:
            raise RevocationManagerError(
                f"No revocation registry record found for id: {rev_reg_id}"
            )

        if notify:
            thread_id = thread_id or f"indy::{rev_reg_id}::{cred_rev_id}"
            rev_notify_rec = RevNotificationRecord(
                rev_reg_id=rev_reg_id,
                cred_rev_id=cred_rev_id,
                thread_id=thread_id,
                connection_id=connection_id,
                comment=comment,
                version=notify_version,
            )
            async with self._profile.session() as session:
                await rev_notify_rec.save(session, reason="New revocation notification")
        if publish:
            await revoc.get_or_fetch_local_tails_path(rev_reg_def)
            result = await revoc.revoke_pending_credentials(
                rev_reg_id,
                additional_crids=[int(cred_rev_id)],
            )

            if result.curr and result.revoked:
                await self.set_cred_revoked_state(rev_reg_id, result.revoked)
                await revoc.update_revocation_list(
                    rev_reg_id,
                    result.prev,
                    result.curr,
                    result.revoked,
                    options=options,
                )

        else:
            await revoc.mark_pending_revocations(rev_reg_id, int(cred_rev_id))

    async def update_rev_reg_revoked_state(
        self,
        rev_reg_def_id: str,
        apply_ledger_update: bool,
        genesis_transactions: str,
    ) -> Tuple[dict, dict, dict]:
        """Request handler to fix ledger entry of credentials revoked against registry.

        This is an indy registry specific operation.

        Args:
            rev_reg_def_id: revocation registry definition id
            apply_ledger_update: whether to apply an update to the ledger
            genesis_transactions: genesis transactions for the ledger

        Returns:
            Number of credentials posted to ledger

        """
        revoc = AnonCredsRevocation(self._profile)
        rev_list = await revoc.get_created_revocation_list(rev_reg_def_id)
        if not rev_list:
            raise RevocationManagerError(
                f"No revocation list found for revocation registry id {rev_reg_def_id}"
            )

        indy_registry = LegacyIndyRegistry()

        if await indy_registry.supports(rev_reg_def_id):
            return await indy_registry.fix_ledger_entry(
                self._profile,
                rev_list,
                apply_ledger_update,
                genesis_transactions,
            )

        raise RevocationManagerError(
            "Indy registry does not support revocation registry "
            f"identified by {rev_reg_def_id}"
        )

    async def publish_pending_revocations(
        self,
        rrid2crid: Optional[Mapping[Text, Sequence[Text]]] = None,
        options: Optional[dict] = None,
    ) -> Mapping[Text, Sequence[Text]]:
        """Publish pending revocations to the ledger.

        Args:
            rrid2crid: Mapping from revocation registry identifiers to all credential
                revocation identifiers within each to publish. Specify null/empty map
                for all revocation registries. Specify empty sequence per revocation
                registry identifier for all pending within the revocation registry;
                e.g.,

                {} - publish all pending revocations from all revocation registries

                {
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0": [],
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1": ["1", "2"]

                } - publish:
                    - all pending revocations from all revocation registry tagged 0
                    - pending ["1", "2"] from revocation registry tagged 1
                    - no pending revocations from any other revocation registries.
            options: Additional options for the revocation registry publish

        Returns: mapping from each revocation registry id to its cred rev ids published.
        """
        options = options or {}
        published_crids = {}
        revoc = AnonCredsRevocation(self._profile)

        rev_reg_def_ids = await revoc.get_revocation_lists_with_pending_revocations()
        for rrid in rev_reg_def_ids:
            if rrid2crid:
                if rrid not in rrid2crid:
                    continue
                limit_crids = [int(crid) for crid in rrid2crid[rrid]]
            else:
                limit_crids = None

            result = await revoc.revoke_pending_credentials(rrid, limit_crids=limit_crids)
            if result.curr and result.revoked:
                await self.set_cred_revoked_state(rrid, result.revoked)
                await revoc.update_revocation_list(
                    rrid, result.prev, result.curr, result.revoked, options
                )
                published_crids[rrid] = sorted(result.revoked)

        return published_crids

    async def clear_pending_revocations(
        self, purge: Mapping[Text, Sequence[Text]] = None
    ) -> Mapping[Text, Sequence[Text]]:
        """Clear pending revocation publications.

        Args:
            purge: Mapping from revocation registry identifiers to all credential
                revocation identifiers within each to clear. Specify null/empty map
                for all revocation registries. Specify empty sequence per revocation
                registry identifier for all pending within the revocation registry;
                e.g.,

                {} - clear all pending revocations from all revocation registries

                {
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0": [],
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1": ["1", "2"]

                } - clear
                    - all pending revocations from all revocation registry tagged 0
                    - pending ["1", "2"] from revocation registry tagged 1
                    - no pending revocations from any other revocation registries.

        Returns:
            mapping from revocation registry id to its remaining
            cred rev ids still marked pending, omitting revocation registries
            with no remaining pending publications.

        """
        result = {}
        notify = []

        revoc = AnonCredsRevocation(self._profile)
        rrids = await revoc.get_revocation_lists_with_pending_revocations()
        async with self._profile.transaction() as txn:
            for rrid in rrids:
                await revoc.clear_pending_revocations(
                    txn,
                    rrid,
                    crid_mask=[int(crid) for crid in (purge or {}).get(rrid, ())],
                )
                remaining = await revoc.get_pending_revocations(rrid)
                if remaining:
                    result[rrid] = remaining
                notify.append(rrid)
            await txn.commit()

        for rrid in notify:
            await notify_pending_cleared_event(self._profile, rrid)

        return result

    async def set_cred_revoked_state(
        self, rev_reg_id: str, cred_rev_ids: Sequence[int]
    ) -> None:
        """Update credentials state to credential_revoked.

        Args:
            rev_reg_id: revocation registry ID
            cred_rev_ids: list of credential revocation IDs

        Returns:
            None

        """
        for cred_rev_id in cred_rev_ids:
            cred_ex_id = None

            try:
                async with self._profile.transaction() as txn:
                    rev_rec = await IssuerCredRevRecord.retrieve_by_ids(
                        txn, rev_reg_id, str(cred_rev_id), for_update=True
                    )
                    cred_ex_id = rev_rec.cred_ex_id
                    cred_ex_version = rev_rec.cred_ex_version
                    rev_rec.state = IssuerCredRevRecord.STATE_REVOKED
                    await rev_rec.save(txn, reason="revoke credential")
                    await txn.commit()
            except StorageNotFoundError:
                continue

            async with self._profile.transaction() as txn:
                if (
                    not cred_ex_version
                    or cred_ex_version == IssuerCredRevRecord.VERSION_1
                ):
                    try:
                        cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                            txn, cred_ex_id, for_update=True
                        )
                        cred_ex_record.state = (
                            V10CredentialExchange.STATE_CREDENTIAL_REVOKED
                        )
                        await cred_ex_record.save(txn, reason="revoke credential")
                        await txn.commit()
                        continue  # skip 2.0 record check
                    except StorageNotFoundError:
                        pass

                if (
                    not cred_ex_version
                    or cred_ex_version == IssuerCredRevRecord.VERSION_2
                ):
                    try:
                        cred_ex_record = await V20CredExRecord.retrieve_by_id(
                            txn, cred_ex_id, for_update=True
                        )
                        cred_ex_record.state = V20CredExRecord.STATE_CREDENTIAL_REVOKED
                        await cred_ex_record.save(txn, reason="revoke credential")
                        await txn.commit()
                    except StorageNotFoundError:
                        pass



"""Model and schema for working with field signatures within message bodies."""

import json
import struct
import time
from typing import Optional, Tuple

from marshmallow import EXCLUDE, fields

from ...protocols.didcomm_prefix import DIDCommPrefix
from ...wallet.base import BaseWallet
from ...wallet.key_type import ED25519
from ...wallet.util import b64_to_bytes, bytes_to_b64
from ..models.base import BaseModel, BaseModelSchema
from ..valid import (
    BASE64URL_EXAMPLE,
    BASE64URL_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
)


class SignatureDecorator(BaseModel):
    """Class representing a field value signed by a known verkey."""

    class Meta:
        """SignatureDecorator metadata."""

        schema_class = "SignatureDecoratorSchema"

    TYPE_ED25519SHA512 = "signature/1.0/ed25519Sha512_single"

    def __init__(
        self,
        *,
        signature_type: Optional[str] = None,
        signature: Optional[str] = None,
        sig_data: Optional[str] = None,
        signer: Optional[str] = None,
    ):
        """Initialize a FieldSignature instance.

        Args:
            signature_type: Type of signature
            signature: The signature
            sig_data: Signature data
            signer: The verkey of the signer
        """
        self.signature_type = signature_type
        self.signature = signature
        self.sig_data = sig_data
        self.signer = signer

    @classmethod
    async def create(
        cls, value, signer: str, wallet: BaseWallet, timestamp=None
    ) -> "SignatureDecorator":
        """Create a Signature.

        Sign a field value and return a newly constructed `SignatureDecorator`
        representing the resulting signature.

        Args:
            cls: SignatureDecorator class
            value: Value to sign
            signer: Verkey of the signing party
            wallet: The wallet to use for the signature
            timestamp: Timestamp to use for the signature

        Returns:
            The created `SignatureDecorator` object

        """
        if not timestamp:
            timestamp = time.time()
        if isinstance(value, BaseModel):
            value = value.serialize()
        # 8 byte, big-endian encoded, unsigned int (long)
        timestamp_bin = struct.pack("!Q", int(timestamp))
        msg_combined_bin = timestamp_bin + json.dumps(value).encode("ascii")
        signature_bin = await wallet.sign_message(msg_combined_bin, signer)
        return SignatureDecorator(
            signature_type=DIDCommPrefix.qualify_current(cls.TYPE_ED25519SHA512),
            signature=bytes_to_b64(signature_bin, urlsafe=True),
            sig_data=bytes_to_b64(msg_combined_bin, urlsafe=True),
            signer=signer,
        )

    def decode(self) -> Tuple[object, int]:
        """Decode the signature to its timestamp and value.

        Returns:
            A tuple of (decoded message, timestamp)

        """
        msg_bin = b64_to_bytes(self.sig_data, urlsafe=True)
        (timestamp,) = struct.unpack_from("!Q", msg_bin, 0)
        return (json.loads(msg_bin[8:]), timestamp)

    async def verify(self, wallet: BaseWallet) -> bool:
        """Verify the signature against the signer's public key.

        Args:
            wallet: Wallet to use to verify signature

        Returns:
            True if verification succeeds else False

        """
        if self.signature_type not in [
            prefix.qualify(SignatureDecorator.TYPE_ED25519SHA512)
            for prefix in DIDCommPrefix
        ]:
            return False
        msg_bin = b64_to_bytes(self.sig_data, urlsafe=True)
        sig_bin = b64_to_bytes(self.signature, urlsafe=True)
        return await wallet.verify_message(msg_bin, sig_bin, self.signer, ED25519)

    def __str__(self):
        """Get a string representation of this class."""
        return (
            f"{self.__class__.__name__}"
            + f"(signature_type='{(self.signature_type,)}', "
            + f"signature='{(self.signature,)}', "
            + f"sig_data='{self.sig_data}', signer='{self.signer}')"
        )


class SignatureDecoratorSchema(BaseModelSchema):
    """SignatureDecorator schema."""

    class Meta:
        """SignatureDecoratorSchema metadata."""

        model_class = SignatureDecorator
        unknown = EXCLUDE

    signature_type = fields.Str(
        data_key="@type",
        required=True,
        metadata={
            "description": "Signature type",
            "example": DIDCommPrefix.NEW.qualify("signature/1.0/ed25519Sha512_single"),
        },
    )
    signature = fields.Str(
        required=True,
        validate=BASE64URL_VALIDATE,
        metadata={
            "description": "signature value, base64url-encoded",
            "example": (
                "FpSxSohK3rhn9QhcJStUNRYUvD8OxLuwda3yhzHkWbZ0VxIbI-l4mKOz7AmkMHDj2"
                "IgDEa1-GCFfWXNl96a7Bg=="
            ),
        },
    )
    sig_data = fields.Str(
        required=True,
        validate=BASE64URL_VALIDATE,
        metadata={
            "description": "Signature data, base64url-encoded",
            "example": BASE64URL_EXAMPLE,
        },
    )
    signer = fields.Str(
        required=True,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Signer verification key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import MENU, PROTOCOL_PACKAGE
from ...models.menu_form import MenuForm
from ...models.menu_form_param import MenuFormParam
from ...models.menu_option import MenuOption
from ..menu import Menu


class TestConfig:
    test_menu_message = {
        "title": "Welcome to IIWBook",
        "description": "IIWBook facilitates ...",
        "options": [
            MenuOption(
                **{
                    "name": "search-introductions",
                    "title": "Search introductions",
                    "description": "Your email address must be verified ...",
                    "disabled": True,
                    "form": MenuForm(
                        **{
                            "title": "Search introductions",
                            "description": "Enter a participant name below ...",
                            "params": [
                                MenuFormParam(
                                    **{
                                        "name": "query",
                                        "title": "Participant name",
                                        "default": "",
                                        "description": "",
                                        "required": True,
                                        "input_type": "text",
                                    }
                                )
                            ],
                            "submit_label": "Search",
                        }
                    ),
                }
            )
        ],
    }


class TestMenu(TestCase, TestConfig):
    def setUp(self):
        self.menu = Menu(**self.test_menu_message)

    def test_init(self):
        """Test initialization."""
        assert self.menu.title == self.test_menu_message["title"]
        assert self.menu.description == self.test_menu_message["description"]
        assert len(self.menu.options) == len(self.test_menu_message["options"])

    def test_type(self):
        """Test type."""
        assert self.menu._type == DIDCommPrefix.qualify_current(MENU)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.menu.MenuSchema.load")
    def test_deserialize(self, mock_menu_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        menu = Menu.deserialize(obj)
        mock_menu_schema_load.assert_called_once_with(obj)

        assert menu is mock_menu_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.menu.MenuSchema.dump")
    def test_serialize(self, mock_menu_schema_dump):
        """
        Test serialization.
        """
        menu_dict = self.menu.serialize()
        mock_menu_schema_dump.assert_called_once_with(self.menu)
        assert menu_dict is mock_menu_schema_dump.return_value

    def test_make_model(self):
        data = self.menu.serialize()
        model_instance = Menu.deserialize(data)
        assert isinstance(model_instance, type(self.menu))



"""Tests for Merkel Validation Utils."""

from unittest import TestCase

from ..utils import ascii_chr, encode_hex


class TestUtils(TestCase):
    """Merkel Validation Utils Tests"""

    def test_encode_hex(self):
        assert encode_hex("test")
        with self.assertRaises(TypeError):
            encode_hex(123)

    def test_aschii_chr(self):
        assert ascii_chr(16 * 5 + 6)



import json
from unittest import IsolatedAsyncioTestCase

from marshmallow import EXCLUDE, fields

from ....cache.base import BaseCache
from ....core.event_bus import Event, EventBus, MockEventBus
from ....messaging.models.base import BaseModelError
from ....storage.base import (
    DEFAULT_PAGE_SIZE,
    BaseStorage,
    StorageDuplicateError,
    StorageRecord,
)
from ....tests import mock
from ....utils.testing import create_test_profile
from ...util import time_now
from ..base_record import BaseRecord, BaseRecordSchema


class BaseRecordImpl(BaseRecord):
    class Meta:
        schema_class = "BaseRecordImplSchema"

    RECORD_TYPE = "record"


class BaseRecordImplSchema(BaseRecordSchema):
    class Meta:
        model_class = BaseRecordImpl
        unknown = EXCLUDE


class ARecordImpl(BaseRecord):
    class Meta:
        schema_class = "ARecordImplSchema"

    RECORD_TYPE = "a-record"
    RECORD_ID_NAME = "ident"
    TAG_NAMES = {"code"}

    def __init__(self, *, ident=None, a, b, code=None, **kwargs):
        super().__init__(ident, **kwargs)
        self.a = a
        self.b = b
        self.code = code

    @property
    def record_value(self) -> dict:
        return {"a": self.a, "b": self.b}


class ARecordImplSchema(BaseRecordSchema):
    class Meta:
        model_class = BaseRecordImpl
        unknown = EXCLUDE

    ident = fields.Str(attribute="_id")
    a = fields.Str()
    b = fields.Str()
    code = fields.Str()


class UnencTestImpl(BaseRecord):
    TAG_NAMES = {"~a", "~b", "c"}


class TestBaseRecord(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

    def test_init_undef(self):
        with self.assertRaises(TypeError):
            BaseRecord()

    def test_from_storage_values(self):
        record_id = "record_id"
        stored = {"created_at": time_now(), "updated_at": time_now()}
        inst = BaseRecordImpl.from_storage(record_id, stored)
        assert isinstance(inst, BaseRecordImpl)
        assert inst._id == record_id
        assert inst.value == stored

        stored[BaseRecordImpl.RECORD_ID_NAME] = inst._id
        with self.assertRaises(ValueError):
            BaseRecordImpl.from_storage(record_id, stored)

    async def test_post_save_new(self):
        async with self.profile.session() as session:
            record = BaseRecordImpl()
            with mock.patch.object(
                record, "post_save", mock.CoroutineMock()
            ) as post_save:
                await record.save(session, reason="reason", event=True)
                post_save.assert_called_once_with(session, True, None, True)

    async def test_post_save_exist(self):
        async with self.profile.session() as session:
            record = BaseRecordImpl()
            await record.save(session, reason="reason", event=True)
            record._last_state = "new"
            await record.save(session, reason="reason", event=False)

    async def test_cache(self):
        async with self.profile.session() as session:
            assert not await BaseRecordImpl.get_cached_key(None, None)
            await BaseRecordImpl.set_cached_key(None, None, None)
            await BaseRecordImpl.clear_cached_key(None, None)
            mock_cache = mock.MagicMock(BaseCache, autospec=True)
            session.context.injector.bind_instance(BaseCache, mock_cache)
            record = BaseRecordImpl()
            cache_key = "cache_key"
            cache_result = await BaseRecordImpl.get_cached_key(session, cache_key)
            mock_cache.get.assert_awaited_once_with(cache_key)
            assert cache_result is mock_cache.get.return_value

            await record.set_cached_key(session, cache_key, record)
            mock_cache.set.assert_awaited_once_with(
                cache_key, record, record.DEFAULT_CACHE_TTL
            )

            await record.clear_cached_key(session, cache_key)
            mock_cache.clear.assert_awaited_once_with(cache_key)

    async def test_retrieve_by_tag_filter_multi_x_delete(self):
        async with self.profile.session() as session:
            records = []
            for i in range(3):
                records.append(ARecordImpl(a="1", b=str(i), code="one"))
                await records[i].save(session)
            with self.assertRaises(StorageDuplicateError):
                await ARecordImpl.retrieve_by_tag_filter(
                    session, {"code": "one"}, {"a": "1"}
                )
            await records[0].delete_record(session)

    async def test_save_x(self):
        async with self.profile.session() as session:
            rec = ARecordImpl(a="1", b="0", code="one")
            with mock.patch.object(session, "inject", mock.MagicMock()) as mock_inject:
                mock_inject.return_value = mock.MagicMock(
                    add_record=mock.CoroutineMock(side_effect=ZeroDivisionError())
                )
                with self.assertRaises(ZeroDivisionError):
                    await rec.save(session)

    async def test_neq(self):
        a_rec = ARecordImpl(a="1", b="0", code="one")
        b_rec = BaseRecordImpl()
        assert a_rec != b_rec

    async def test_query(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            record_value = {"created_at": time_now(), "updated_at": time_now()}
            tag_filter = {"tag": "filter"}
            stored = StorageRecord(
                BaseRecordImpl.RECORD_TYPE, json.dumps(record_value), {}, record_id
            )

            mock_storage.find_all_records.return_value = [stored]
            result = await BaseRecordImpl.query(session, tag_filter)
            mock_storage.find_all_records.assert_awaited_once_with(
                type_filter=BaseRecordImpl.RECORD_TYPE,
                order_by=None,
                descending=False,
                tag_query=tag_filter,
            )
            assert result and isinstance(result[0], BaseRecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value

    async def test_query_x(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            record_value = {"created_at": time_now(), "updated_at": time_now()}
            tag_filter = {"tag": "filter"}
            stored = StorageRecord(
                BaseRecordImpl.RECORD_TYPE, json.dumps(record_value), {}, record_id
            )

            mock_storage.find_all_records.return_value = [stored]
            with mock.patch.object(
                BaseRecordImpl,
                "from_storage",
                mock.MagicMock(side_effect=BaseModelError),
            ):
                with self.assertRaises(BaseModelError):
                    await BaseRecordImpl.query(session, tag_filter)

    async def test_query_post_filter(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            a_record = ARecordImpl(ident=record_id, a="one", b="two", code="red")
            record_value = a_record.record_value
            record_value.update({"created_at": time_now(), "updated_at": time_now()})
            tag_filter = {"code": "red"}
            post_filter_pos_alt = {"a": ["one", "a suffusion of yellow"]}
            post_filter_neg_alt = {"a": ["three", "no, five"]}
            stored = StorageRecord(
                ARecordImpl.RECORD_TYPE,
                json.dumps(record_value),
                {"code": "red"},
                record_id,
            )
            mock_storage.find_all_records.return_value = [stored]

            # positive match
            result = await ARecordImpl.query(
                session, tag_filter, post_filter_positive={"a": "one"}
            )
            mock_storage.find_all_records.assert_awaited_once_with(
                type_filter=ARecordImpl.RECORD_TYPE,
                order_by=None,
                descending=False,
                tag_query=tag_filter,
            )
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"

            # positive match by list of alternatives to hit
            result = await ARecordImpl.query(
                session, tag_filter, post_filter_positive=post_filter_pos_alt, alt=True
            )
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"

            # negative match by list of alternatives to miss (result complies)
            result = await ARecordImpl.query(
                session, tag_filter, post_filter_negative=post_filter_neg_alt, alt=True
            )
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"

            # negative match by list of alternatives to miss, with one hit spoiling result
            post_filter_neg_alt = {"a": ["one", "three", "no, five"]}
            result = await ARecordImpl.query(
                session, tag_filter, post_filter_negative=post_filter_neg_alt, alt=True
            )
            assert not result

    @mock.patch("builtins.print")
    def test_log_state(self, mock_print):
        test_param = "test.log"
        with mock.patch.object(BaseRecordImpl, "LOG_STATE_FLAG", test_param):
            record = BaseRecordImpl()
            record.log_state(
                msg="state",
                params={"a": "1", "b": "2"},
                settings=mock.MagicMock(get={test_param: 1}.get),
            )
        mock_print.assert_called_once()

    @mock.patch("builtins.print")
    def test_skip_log(self, mock_print):
        record = BaseRecordImpl()
        record.log_state("state", settings=None)
        mock_print.assert_not_called()

    async def test_emit_event(self):
        async with self.profile.session() as session:
            mock_event_bus = MockEventBus()
            session.profile.context.injector.bind_instance(EventBus, mock_event_bus)
            record = BaseRecordImpl()
            payload = {"test": "payload"}

            # Records must have topic to emit events
            record.RECORD_TOPIC = None
            await record.emit_event(session, payload)
            assert mock_event_bus.events == []

            record.RECORD_TOPIC = "topic"

            # Stateless record with no payload emits event with serialized record
            await record.emit_event(session)
            assert mock_event_bus.events == [
                (session.profile, Event("acapy::record::topic", {}))
            ]
            mock_event_bus.events.clear()

            # Stateless record with payload emits event
            await record.emit_event(session, payload)
            assert mock_event_bus.events == [
                (session.profile, Event("acapy::record::topic", payload))
            ]
            mock_event_bus.events.clear()

            # Statefull record with payload emits event
            record.state = "test_state"
            await record.emit_event(session, payload)
            assert mock_event_bus.events == [
                (session.profile, Event("acapy::record::topic::test_state", payload))
            ]

    async def test_tag_prefix(self):
        tags = {"~x": "a", "y": "b"}
        assert UnencTestImpl.strip_tag_prefix(tags) == {"x": "a", "y": "b"}

        tags = {"a": "x", "b": "y", "c": "z"}
        assert UnencTestImpl.prefix_tag_filter(tags) == {"~a": "x", "~b": "y", "c": "z"}

        tags = {"$not": {"a": "x", "b": "y", "c": "z"}}
        expect = {"$not": {"~a": "x", "~b": "y", "c": "z"}}
        actual = UnencTestImpl.prefix_tag_filter(tags)
        assert {**expect} == {**actual}

        tags = {"$or": [{"a": "x"}, {"c": "z"}]}
        assert UnencTestImpl.prefix_tag_filter(tags) == {"$or": [{"~a": "x"}, {"c": "z"}]}

    async def test_query_with_limit(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            a_record = ARecordImpl(ident=record_id, a="one", b="two", code="red")
            record_value = a_record.record_value
            record_value.update({"created_at": time_now(), "updated_at": time_now()})
            tag_filter = {"code": "red"}
            stored = StorageRecord(
                ARecordImpl.RECORD_TYPE,
                json.dumps(record_value),
                {"code": "red"},
                record_id,
            )
            mock_storage.find_paginated_records.return_value = [stored]

            # Query with limit
            result = await ARecordImpl.query(session, tag_filter, limit=10)
            mock_storage.find_paginated_records.assert_awaited_once_with(
                type_filter=ARecordImpl.RECORD_TYPE,
                tag_query=tag_filter,
                limit=10,
                offset=0,
                order_by=None,
                descending=False,
            )
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"

    async def test_query_with_offset(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            a_record = ARecordImpl(ident=record_id, a="one", b="two", code="red")
            record_value = a_record.record_value
            record_value.update({"created_at": time_now(), "updated_at": time_now()})
            tag_filter = {"code": "red"}
            stored = StorageRecord(
                ARecordImpl.RECORD_TYPE,
                json.dumps(record_value),
                {"code": "red"},
                record_id,
            )
            mock_storage.find_paginated_records.return_value = [stored]

            # Query with offset
            result = await ARecordImpl.query(session, tag_filter, offset=10)
            mock_storage.find_paginated_records.assert_awaited_once_with(
                type_filter=ARecordImpl.RECORD_TYPE,
                tag_query=tag_filter,
                limit=DEFAULT_PAGE_SIZE,
                offset=10,
                order_by=None,
                descending=False,
            )
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"

    async def test_query_with_limit_and_offset(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            a_record = ARecordImpl(ident=record_id, a="one", b="two", code="red")
            record_value = a_record.record_value
            record_value.update({"created_at": time_now(), "updated_at": time_now()})
            tag_filter = {"code": "red"}
            stored = StorageRecord(
                ARecordImpl.RECORD_TYPE,
                json.dumps(record_value),
                {"code": "red"},
                record_id,
            )
            mock_storage.find_paginated_records.return_value = [stored]

            # Query with limit and offset
            result = await ARecordImpl.query(session, tag_filter, limit=10, offset=5)
            mock_storage.find_paginated_records.assert_awaited_once_with(
                type_filter=ARecordImpl.RECORD_TYPE,
                tag_query=tag_filter,
                limit=10,
                offset=5,
                order_by=None,
                descending=False,
            )
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"

    async def test_query_with_limit_and_offset_and_post_filter(self):
        async with self.profile.session() as session:
            mock_storage = mock.MagicMock(BaseStorage, autospec=True)
            session.context.injector.bind_instance(BaseStorage, mock_storage)
            record_id = "record_id"
            a_record = ARecordImpl(ident=record_id, a="one", b="two", code="red")
            record_value = a_record.record_value
            record_value.update({"created_at": time_now(), "updated_at": time_now()})
            tag_filter = {"code": "red"}
            stored = StorageRecord(
                ARecordImpl.RECORD_TYPE,
                json.dumps(record_value),
                {"code": "red"},
                record_id,
            )
            mock_storage.find_all_records.return_value = [
                stored
            ] * 15  # return 15 records

            # Query with limit and offset
            result = await ARecordImpl.query(
                session,
                tag_filter,
                limit=10,
                offset=5,
                post_filter_positive={"a": "one"},
            )
            mock_storage.find_all_records.assert_awaited_once_with(
                type_filter=ARecordImpl.RECORD_TYPE,
                order_by=None,
                descending=False,
                tag_query=tag_filter,
            )
            assert len(result) == 10
            assert result and isinstance(result[0], ARecordImpl)
            assert result[0]._id == record_id
            assert result[0].value == record_value
            assert result[0].a == "one"



"""DataIntegrityProof."""

from typing import Optional

from marshmallow import INCLUDE, fields, post_dump

from ....messaging.models.base import BaseModel, BaseModelSchema
from ....messaging.valid import RFC3339_DATETIME_EXAMPLE, UUID4_EXAMPLE, Uri


class DataIntegrityProofOptions(BaseModel):
    """Data Integrity Proof Options model."""

    class Meta:
        """DataIntegrityProofOptions metadata."""

        schema_class = "DataIntegrityProofOptionsSchema"

    def __init__(
        self,
        id: Optional[str] = None,
        type: Optional[str] = None,
        proof_purpose: Optional[str] = None,
        verification_method: Optional[str] = None,
        cryptosuite: Optional[str] = None,
        created: Optional[str] = None,
        expires: Optional[str] = None,
        domain: Optional[str] = None,
        challenge: Optional[str] = None,
        previous_proof: Optional[str] = None,
        proof_value: Optional[str] = None,
        nonce: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Initialize the DataIntegrityProofOptions instance."""

        self.id = id
        self.type = type
        self.proof_purpose = proof_purpose
        self.verification_method = verification_method
        self.cryptosuite = cryptosuite
        self.created = created
        self.expires = expires
        self.domain = domain
        self.challenge = challenge
        self.previous_proof = previous_proof
        self.proof_value = proof_value
        self.nonce = nonce
        self.extra = kwargs


class DataIntegrityProofOptionsSchema(BaseModelSchema):
    """Data Integrity Proof Options schema.

    Based on https://www.w3.org/TR/vc-data-integrity/#proofs

    """

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = DataIntegrityProofOptions

    id = fields.Str(
        required=False,
        metadata={
            "description": (
                "An optional identifier for the proof, which MUST be a URL [URL], "
                "such as a UUID as a URN"
            ),
            "example": "urn:uuid:6a1676b8-b51f-11ed-937b-d76685a20ff5",
        },
    )

    type = fields.Str(
        required=True,
        metadata={
            "description": (
                "The specific type of proof MUST be specified as a string that maps "
                "to a URL [URL]."
            ),
            "example": "DataIntegrityProof",
        },
    )

    proof_purpose = fields.Str(
        data_key="proofPurpose",
        required=True,
        metadata={
            "description": (
                "The proof purpose acts as a safeguard to prevent the "
                "proof from being misused by being applied to a purpose other than "
                "the one that was intended."
            ),
            "example": "assertionMethod",
        },
    )

    verification_method = fields.Str(
        data_key="verificationMethod",
        required=True,
        validate=Uri(),
        metadata={
            "description": (
                "A verification method is the means and information "
                "needed to verify the proof."
            ),
            "example": (
                "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg34"
                "2Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
            ),
        },
    )

    cryptosuite = fields.Str(
        required=True,
        metadata={
            "description": (
                "An identifier for the cryptographic suite that can be used to "
                "verify the proof."
            ),
            "example": "eddsa-jcs-2022",
        },
    )

    created = fields.Str(
        required=False,
        metadata={
            "description": (
                "The date and time the proof was created is OPTIONAL and, if "
                "included, MUST be specified as an [XMLSCHEMA11-2] "
                "dateTimeStamp string"
            ),
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )

    expires = fields.Str(
        required=False,
        metadata={
            "description": (
                "The expires property is OPTIONAL and, if present, specifies when "
                "the proof expires. If present, it MUST be an [XMLSCHEMA11-2] "
                "dateTimeStamp string"
            ),
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )

    domain = fields.Str(
        required=False,
        metadata={
            "description": (
                "It conveys one or more security domains in which the proof is "
                "meant to be used."
            ),
            "example": "example.com",
        },
    )

    challenge = fields.Str(
        required=False,
        metadata={
            "description": (
                "The value is used once for a particular domain and window of time. "
                "This value is used to mitigate replay attacks."
            ),
            "example": UUID4_EXAMPLE,
        },
    )

    previous_proof = fields.Str(
        required=False,
        data_key="previousProof",
        metadata={
            "description": (
                "Each value identifies another data integrity proof that "
                "MUST verify before the current proof is processed."
            ),
            "example": ("urn:uuid:6a1676b8-b51f-11ed-937b-d76685a20ff5"),
        },
    )

    proof_value = fields.Str(
        required=False,
        data_key="proofValue",
        metadata={
            "description": "The value of the proof signature.",
            "example": (
                "zsy1AahqbzJQ63n9RtekmwzqZeVj494VppdAVJBnMYrTwft6cLJJGeTSSxCCJ6HKnR"
                "twE7jjDh6sB2z2AAiZY9BBnCD8wUVgwqH3qchGRCuC2RugA4eQ9fUrR4Yuycac3caiaaay"
            ),
        },
    )

    nonce = fields.Str(
        required=False,
        metadata={
            "description": (
                "One use of this field is to increase privacy by decreasing linkability "
                "that is the result of deterministically generated signatures."
            ),
            "example": (
                "CF69iO3nfvqRsRBNElE8b4wO39SyJHPM7Gg1nExltW5vSfQA1lvDCR/zXX1To0/4NLo=="
            ),
        },
    )

    @post_dump(pass_original=True)
    def add_unknown_properties(self, data: dict, original, **kwargs):
        """Add back unknown properties before outputting."""

        data.update(original.extra)

        return data



"""Handshake Reuse Message Handler under RFC 0434."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..manager import OutOfBandManager, OutOfBandManagerError
from ..messages.reuse import HandshakeReuse


class HandshakeReuseMessageHandler(BaseHandler):
    """Handler class for Handshake Reuse Message Handler under RFC 0434."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle Handshake Reuse Message Handler under RFC 0434.

        Args:
            context: Request context
            responder: Responder callback
        """
        self._logger.debug("HandshakeReuseMessageHandler called with context %s", context)
        assert isinstance(context.message, HandshakeReuse)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        profile = context.profile
        mgr = OutOfBandManager(profile)
        try:
            await mgr.receive_reuse_message(
                context.message, context.message_receipt, context.connection_record
            )
        except OutOfBandManagerError as e:
            self._logger.exception(f"Error processing Handshake Reuse message, {e}")



"""Test for DIF PresExch Schemas."""

import json
from unittest import TestCase

from .....messaging.models.base import BaseModelError
from ..pres_exch import (
    ClaimFormat,
    Constraints,
    DIFHolder,
    Filter,
    SchemasInputDescriptorFilter,
    SubmissionRequirements,
    VPWithSubmission,
)


class TestPresExchSchemas(TestCase):
    """Presentation predicate specification tests"""

    def test_claim_format_a(self):
        submission_req_json = """
            {
                "jwt": {
                "alg": ["EdDSA", "ES256K", "ES384"]
                },
                "jwt_vc": {
                "alg": ["ES256K", "ES384"]
                },
                "jwt_vp": {
                "alg": ["EdDSA", "ES256K"]
                },
                "ldp_vc": {
                "proof_type": [
                    "JsonWebSignature2020",
                    "Ed25519Signature2018",
                    "EcdsaSecp256k1Signature2019",
                    "RsaSignature2018"
                ]
                },
                "ldp_vp": {
                "proof_type": ["Ed25519Signature2018"]
                },
                "ldp": {
                "proof_type": ["RsaSignature2018"]
                }
            }
        """
        expected_result = json.loads(submission_req_json)
        actual_result = (ClaimFormat.deserialize(submission_req_json)).serialize()
        assert expected_result == actual_result

    def test_claim_format_b(self):
        submission_req_json = """
            {
                "ldp_vp": {
                "proof_type": ["Ed25519Signature2018"]
                }
            }
        """
        expected_result = json.loads(submission_req_json)
        actual_result = (ClaimFormat.deserialize(submission_req_json)).serialize()
        assert expected_result == actual_result

    def test_submission_requirements_from(self):
        claim_format_json = """
            {
                "name": "European Union Citizenship Proofs",
                "rule": "pick",
                "min": 1,
                "from": "A"
            }
        """
        expected_result = json.loads(claim_format_json)
        actual_result = (
            SubmissionRequirements.deserialize(claim_format_json)
        ).serialize()
        assert expected_result == actual_result

    def test_submission_requirements_from_nested(self):
        nested_submission_req_json = """
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "count": 1,
                "from_nested": [
                    {
                    "name": "United States Citizenship Proofs",
                    "purpose": "We need you to prove you are a US citizen.",
                    "rule": "all",
                    "from": "A"
                    },
                    {
                    "name": "European Union Citizenship Proofs",
                    "purpose": "We need you to prove you are a citizen of a EU country.",
                    "rule": "all",
                    "from": "B"
                    }
                ]
            }
        """
        expected_result = json.loads(nested_submission_req_json)
        actual_result = (
            SubmissionRequirements.deserialize(nested_submission_req_json)
        ).serialize()
        assert expected_result == actual_result

    def test_submission_requirements_from_nested_of_nested(self):
        nested_submission_req_json = """
            {
                "name": "Лабораторијски резултати",
                "purpose": "Morbilli virus критеријум",
                "rule": "pick",
                "count": 1,
                "from_nested": [
                    {
                        "name": "Лични подаци",
                        "purpose": "Морамо идентификовати субјекта акредитива",
                        "rule": "pick",
                        "count": 1,
                        "from": "Patient"
                    },
                    {
                        "name": "Лабораторијски резултати",
                        "purpose": "Morbilli virus критеријум",
                        "rule": "pick",
                        "count": 1,
                        "from_nested": [
                            {
                                "name": "Негативан тест у последња 24 сата",
                                "purpose": "Незаразност",
                                "rule": "pick",
                                "count": 1,
                                "from": "PCR"
                            },
                            {
                                "name": "Тест атнитела",
                                "purpose": "Имуност",
                                "rule": "pick",
                                "count": 1,
                                "from": "IgG"
                            }
                        ]
                    }
                ]
            }
        """
        expected_result = json.loads(nested_submission_req_json)
        assert SubmissionRequirements.deserialize(nested_submission_req_json)
        actual_result = (
            SubmissionRequirements.deserialize(nested_submission_req_json)
        ).serialize()
        assert expected_result == actual_result

    def test_submission_requirements_from_missing(self):
        test_json = """
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "count": 1
            }
        """
        with self.assertRaises(BaseModelError) as cm:
            (SubmissionRequirements.deserialize(test_json)).serialize()

    def test_submission_requirements_from_both_present(self):
        test_json = """
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "count": 1,
                "from": "A",
                "from_nested": [
                    {
                    "name": "United States Citizenship Proofs",
                    "purpose": "We need you to prove you are a US citizen.",
                    "rule": "all",
                    "from": "A"
                    },
                    {
                    "name": "European Union Citizenship Proofs",
                    "purpose": "We need you to prove you are a citizen of a EU country.",
                    "rule": "all",
                    "from": "B"
                    }
                ]
            }
        """
        with self.assertRaises(BaseModelError) as cm:
            (SubmissionRequirements.deserialize(test_json)).serialize()

    def test_submission_requirements_from_both_missing(self):
        test_json = """
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "count": 1
            }
        """
        with self.assertRaises(BaseModelError) as cm:
            (SubmissionRequirements.deserialize(test_json)).serialize()

    def test_is_holder(self):
        test_json = """
            {
                "field_id": [
                    "ce66380c-1990-4aec-b8b4-5d532e92a616",
                    "dd69e8a4-4cc0-4540-b34a-b4aa0e0d2214",
                    "d15802b4-eec8-45ef-b78f-e35125ac1bb8",
                    "765f3e09-600c-467f-99eb-ea549c350121"
                ],
                "directive": "required"
            }
        """
        expected_result = json.loads(test_json)
        actual_result = (DIFHolder.deserialize(test_json)).serialize()
        assert expected_result == actual_result

    def test_filter(self):
        test_json_list = []
        test_json_string_enum = """
            {
              "type":"string",
              "enum": ["testa1", "testa2", "testa3"]
            }
        """
        test_json_list.append(test_json_string_enum)
        test_json_number_enum = """
            {
              "type":"string",
              "enum": ["testb1", "testb2", "testb3"]
            }
        """
        test_json_list.append(test_json_number_enum)
        test_json_not_enum = """
            {
                "not": {
                    "enum": ["testc1", "testc2", "testc3"]
                }
            }
        """
        test_json_list.append(test_json_not_enum)
        test_json_format_min = """
            {
                "type":"string",
                "format": "date",
                "minimum": "1980/07/04" 
            }
        """
        test_json_list.append(test_json_format_min)
        test_json_exclmax = """
            {
                "type":"number",
                "exclusiveMaximum": 2
            }
        """
        test_json_list.append(test_json_exclmax)
        test_json_exclmin = """
            {
                "exclusiveMinimum": 2
            }
        """
        test_json_list.append(test_json_exclmin)
        test_json_const = """
            {
                "const": 2.0
            }
        """
        test_json_list.append(test_json_const)
        test_json_enum_error = """
            {
                "enum": 2
            }
        """
        test_json_custom_field_error = """
            {
                "minimum": [
                    "not_valid"
                ]
            }
        """

        for tmp_test_item in test_json_list:
            expected_result = json.loads(tmp_test_item)
            actual_result = (Filter.deserialize(tmp_test_item)).serialize()
            assert expected_result == actual_result

        with self.assertRaises(BaseModelError) as cm:
            (Filter.deserialize(test_json_enum_error)).serialize()

        with self.assertRaises(BaseModelError) as cm:
            (Filter.deserialize(test_json_custom_field_error)).serialize()

    def test_constraints(self):
        test_json = """
            {
                "fields":[
                    {
                        "path":[
                            "$.credentialSubject.dob",
                            "$.vc.credentialSubject.dob",
                            "$.credentialSubject.license.dob"
                        ],
                        "filter":{
                        "type":"string",
                        "format":"date",
                        "minimum":"1999-5-16"
                        }
                    }
                ],
                "statuses": {
                    "active": {
                        "directive": "required"
                    },
                    "suspended": {
                        "directive": "allowed"
                    },
                    "revoked": {
                        "directive": "disallowed"
                    }
                }
            }
        """

        expected_result = json.loads(test_json)
        actual_result = (Constraints.deserialize(test_json)).serialize()
        assert expected_result == actual_result

    def test_verifiable_presentation_wrapper(self):
        test_vp_dict = {
            "@context": ["https://www.w3.org/2018/credentials/v1"],
            "type": ["VerifiablePresentation"],
            "verifiableCredential": [
                {
                    "@context": [
                        "https://www.w3.org/2018/credentials/v1",
                        "https://w3id.org/citizenship/v1",
                        "https://w3id.org/security/bbs/v1",
                    ],
                    "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
                    "type": ["PermanentResidentCard", "VerifiableCredential"],
                    "credentialSubject": {
                        "id": "did:example:b34ca6cd37bbf23",
                        "type": ["Person", "PermanentResident"],
                        "givenName": "JOHN",
                    },
                    "issuanceDate": "2010-01-01T19:53:24Z",
                    "issuer": "did:key:zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v",
                    "proof": {
                        "type": "BbsBlsSignatureProof2020",
                        "nonce": "3AuruhJQrXtEgiagiJ+FwVf2S0SnzUDJvnO61YecQsJ7ImR1mPcoVjJJ0HOhfkFpoYI=",
                        "proofValue": "ABkBuAaPlP5A7JWY78Xf69oBnsMLcD1RXbIFYhcLoXPXW12CG9glnnqnPLsGri5xsA3LcP0kg74X+sAjKXGRGy3uvp412Dm0FuohYNboQcLne5KOAa5AxU4bjmwQsxdfduVqhriro1N+YTkuB4SMmO/5ooL0N3OHsYdExg7nSzWqmZoqgp+3CwIxF0a/oyKTcxJORuIqAAAAdInlL9teSIX49NJGEZfBO7IrdjT2iggH/G0AlPWoEvrWIbuCRQ69K83n5o7oJVjqhAAAAAIaVmlAD6+FEKA4eg0OaWOKPrd5Kq8rv0vIwjJ71egxll0Fqq4zDWQ/+yl3Pteh0Wyuyvpm19/sj6tiCWj4PkA+rpxtR2bXpnrCTKUffFFNBjVvVziXDS0KWkGUB7XU9mjUa4USC7Iub3bZZCnFjQA5AAAADzkGwGD837r33e7OTrGEti8eAkvFDcyCgA4ck/X+5HJjAJclHWbl4SNQR8CiNZyzJpvxW+jbNBcwmEvocYArddk3F78Ki0Qnp6aU9eDgfOOx1iW2BXLUjrhq5I2hP5/WQF3CEDYRjczGjzM9T8/coeC36YAp0zJunIXUKb8SPDSOISafibYRYFB4xhlWKXWloDelafyujOBST8KZNM8FmF4DSbXrO8vmZbjuR/8ntUcUK7X2rNbuZ3M5eWZDF8pL+SA9gQitKfPHEocoYAdhgEAM7ZNAJ+TgOcx9gtZIhDWKDNnFxIeoOAylbD1xZd9xbWtq3Bk3R79xqsKxFRJRNxk/9b6fJruP292+qM5lxcZ1jUz/dJUYFI93hH4Mso75CjGRN78MAY9SNifl6H8qcxTpBn4332LlFhRznLbtnc4YSWA/fvVqaN9h2zCH/6AdbLKXGffV34EF7DadwJsi9jsc+YlSMn6qaIUIDTdGLwh4KKpSH5bVbg/mVCcXPTJplFgYwRsOdiQbZY/740dJyo1lPjQ0Lvdio8W2M8c73ujeJU70CNLkgjJAMUPGrCFtGxBH2eeLBQ0P95qRZAIcJ7U0MibZLaRjoUOuTla5BIt2038PJ6XhcY6BEJaLyJOPEQ==",
                        "verificationMethod": "did:key:zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v#zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v",
                        "proofPurpose": "assertionMethod",
                        "created": "2021-05-05T15:22:30.523465",
                    },
                }
            ],
            "presentation_submission": {
                "id": "a5fcfe44-2c30-497d-af02-98e539da9a0f",
                "definition_id": "32f54163-7166-48f1-93d8-ff217bdb0653",
                "descriptor_map": [
                    {
                        "id": "citizenship_input_1",
                        "format": "ldp_vc",
                        "path": "$.verifiableCredential[0]",
                    }
                ],
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:sov:4QxzWk3ajdnEA37NdNU5Kt#key-1",
                "created": "2021-05-05T15:23:03.023971",
                "proofPurpose": "authentication",
                "challenge": "40429d49-5e8f-4ffc-baf8-e332412f1247",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..2uBYmg7muE9ZPVeAGo_ibVfLkCjf2hGshr2o5i8pAwFyNBM-kDHXofuq1MzJgb19wzb01VIu91hY_ajjt9KFAA",
            },
        }
        vp = VPWithSubmission.deserialize(test_vp_dict)
        assert isinstance(vp, VPWithSubmission)

    def test_schemas_input_desc_filter(self):
        test_schema_list = [
            [
                {"uri": "https://www.w3.org/2018/VC"},
                {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
            ],
            [{"uri": "https://www.w3.org/Test#Test"}],
        ]
        test_schemas_filter = {
            "oneof_filter": test_schema_list,
        }

        deser_schema_filter = SchemasInputDescriptorFilter.deserialize(
            test_schemas_filter
        )
        ser_schema_filter = deser_schema_filter.serialize()
        deser_schema_filter = SchemasInputDescriptorFilter.deserialize(ser_schema_filter)
        assert deser_schema_filter.oneof_filter
        assert deser_schema_filter.uri_groups[0][0].uri == test_schema_list[0][0].get(
            "uri"
        )
        assert deser_schema_filter.uri_groups[0][1].uri == test_schema_list[0][1].get(
            "uri"
        )
        assert deser_schema_filter.uri_groups[1][0].uri == test_schema_list[1][0].get(
            "uri"
        )
        assert isinstance(deser_schema_filter, SchemasInputDescriptorFilter)

        test_schema_list = [
            {"uri": "https://www.w3.org/Test#Test"},
            {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
        ]
        test_schemas_filter = {
            "oneof_filter": test_schema_list,
        }

        deser_schema_filter = SchemasInputDescriptorFilter.deserialize(
            test_schemas_filter
        )
        ser_schema_filter = deser_schema_filter.serialize()
        deser_schema_filter = SchemasInputDescriptorFilter.deserialize(ser_schema_filter)
        assert deser_schema_filter.oneof_filter
        assert deser_schema_filter.uri_groups[0][0].uri == test_schema_list[0].get("uri")
        assert deser_schema_filter.uri_groups[1][0].uri == test_schema_list[1].get("uri")
        assert isinstance(deser_schema_filter, SchemasInputDescriptorFilter)

        deser_schema_filter = SchemasInputDescriptorFilter.deserialize(test_schema_list)
        ser_schema_filter = deser_schema_filter.serialize()
        deser_schema_filter = SchemasInputDescriptorFilter.deserialize(ser_schema_filter)
        assert not deser_schema_filter.oneof_filter
        assert deser_schema_filter.uri_groups[0][0].uri == test_schema_list[0].get("uri")
        assert deser_schema_filter.uri_groups[0][1].uri == test_schema_list[1].get("uri")
        assert isinstance(deser_schema_filter, SchemasInputDescriptorFilter)



"""Revoke message."""

from typing import Optional

from marshmallow import fields, validate

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.please_ack_decorator import (
    PleaseAckDecorator,
    PleaseAckDecoratorSchema,
)
from .....messaging.valid import UUID4_EXAMPLE
from ..message_types import PROTOCOL_PACKAGE, REVOKE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.revoke_handler.RevokeHandler"


class Revoke(AgentMessage):
    """Class representing revoke message."""

    class Meta:
        """Revoke Meta."""

        handler_class = HANDLER_CLASS
        message_type = REVOKE
        schema_class = "RevokeSchema"

    def __init__(
        self,
        *,
        revocation_format: str,
        credential_id: str,
        please_ack: Optional[PleaseAckDecorator] = None,
        comment: Optional[str] = None,
        **kwargs,
    ):
        """Initialize revoke message."""
        super().__init__(**kwargs)
        self.revocation_format = revocation_format
        self.credential_id = credential_id
        self.comment = comment


class RevokeSchema(AgentMessageSchema):
    """Schema of Revoke message."""

    class Meta:
        """RevokeSchema Meta."""

        model_class = Revoke

    revocation_format = fields.Str(
        required=True,
        validate=validate.OneOf(["indy-anoncreds"]),
        metadata={
            "description": "The format of the credential revocation ID",
            "example": "indy-anoncreds",
        },
    )
    credential_id = fields.Str(
        required=True,
        metadata={
            "description": "Credential ID of the issued credential to be revoked",
            "example": UUID4_EXAMPLE,
        },
    )
    please_ack = fields.Nested(
        PleaseAckDecoratorSchema,
        required=False,
        data_key="~please_ack",
        metadata={"description": "Whether or not the holder should acknowledge receipt"},
    )
    comment = fields.Str(
        required=False,
        metadata={
            "description": "Human readable information about revocation notification"
        },
    )



"""Generic ack message handler."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....utils.tracing import get_timer, trace_event
from ..messages.ack import V10Ack


class V10AckHandler(BaseHandler):
    """Message handler class for generic acks."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for presentation acks.

        Args:
            context: request context
            responder: responder callback
        """
        r_time = get_timer()

        self._logger.debug("V20PresAckHandler called with context %s", context)
        assert isinstance(context.message, V10Ack)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        self._logger.info(
            "Received v1.0 notification ack message: %s",
            context.message.serialize(as_string=True),
        )

        trace_event(
            context.settings,
            context.message,
            outcome="V10AckHandler.handle.END",
            perf_counter=r_time,
        )



from unittest import IsolatedAsyncioTestCase

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.presentation_proposal import PresentationProposal
from .. import presentation_proposal_handler as test_module


class TestPresentationProposalHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()
        self.request_context = RequestContext.test_context(self.profile)
        self.request_context.message_receipt = MessageReceipt()
        self.request_context.settings["debug.auto_respond_presentation_proposal"] = False
        self.request_context.connection_record = mock.MagicMock()
        self.request_context.message = mock.MagicMock()
        self.request_context.message.comment = "hello world"

    async def test_called(self):
        with mock.patch.object(
            test_module, "PresentationManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_proposal = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            self.request_context.message = PresentationProposal()
            self.request_context.connection_ready = True
            self.request_context.connection_record = mock.MagicMock()
            handler = test_module.PresentationProposalHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)

        mock_pres_mgr.assert_called_once_with(self.request_context.profile)
        mock_pres_mgr.return_value.receive_proposal.assert_called_once_with(
            self.request_context.message, self.request_context.connection_record
        )
        assert not responder.messages

    async def test_called_auto_request(self):
        self.request_context.settings["debug.auto_respond_presentation_proposal"] = True
        with mock.patch.object(
            test_module, "PresentationManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_proposal = mock.CoroutineMock(
                return_value="presentation_exchange_record"
            )
            mock_pres_mgr.return_value.create_bound_request = mock.CoroutineMock(
                return_value=(
                    mock_pres_mgr.return_value.receive_proposal.return_value,
                    "presentation_request_message",
                )
            )
            self.request_context.message = PresentationProposal()
            self.request_context.connection_ready = True
            handler = test_module.PresentationProposalHandler()
            responder = MockResponder()
            await handler.handle(self.request_context, responder)

        mock_pres_mgr.assert_called_once_with(self.request_context.profile)
        mock_pres_mgr.return_value.create_bound_request.assert_called_once_with(
            presentation_exchange_record=(
                mock_pres_mgr.return_value.receive_proposal.return_value
            ),
            comment=self.request_context.message.comment,
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "presentation_request_message"
        assert target == {}

    async def test_called_auto_request_x(self):
        with mock.patch.object(
            test_module, "PresentationManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_proposal = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_pres_mgr.return_value.create_bound_request = mock.CoroutineMock(
                side_effect=test_module.LedgerError()
            )

            self.request_context.message = PresentationProposal()
            self.request_context.connection_ready = True
            handler = test_module.PresentationProposalHandler()
            responder = MockResponder()

            await handler.handle(self.request_context, responder)

    async def test_called_not_ready(self):
        with mock.patch.object(
            test_module, "PresentationManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_proposal = mock.CoroutineMock()
            self.request_context.message = PresentationProposal()
            self.request_context.connection_ready = False
            handler = test_module.PresentationProposalHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(self.request_context, responder)
        assert (
            err.exception.message == "Connection used for presentation proposal not ready"
        )

        assert not responder.messages

    async def test_called_no_connection(self):
        self.request_context.message = PresentationProposal()
        self.request_context.connection_record = None
        handler = test_module.PresentationProposalHandler()
        responder = MockResponder()
        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(self.request_context, responder)
        assert (
            err.exception.message
            == "Connectionless not supported for presentation proposal"
        )

        assert not responder.messages



from unittest import IsolatedAsyncioTestCase

from marshmallow.exceptions import ValidationError

from ..route_record import RouteRecordSchema


class TestConnRecord(IsolatedAsyncioTestCase):
    async def test_route_record_validation_fails_no_connection_wallet(self):
        schema = schema = RouteRecordSchema()

        with self.assertRaises(ValidationError):
            schema.validate_fields({})

        schema.validate_fields({"connection_id": "dummy"})
        schema.validate_fields({"wallet_id": "dummy"})



import asyncio
from unittest import mock

import pytest

from ...askar.profile import AskarProfile, AskarProfileManager
from ...config.injection_context import InjectionContext
from ...ledger.base import BaseLedger
from .. import profile as test_module
from ..profile_anon import AskarAnonProfileManager


@pytest.fixture
def open_store():
    yield mock.MagicMock()


@pytest.mark.asyncio
async def test_init_success(open_store):
    askar_profile = AskarProfile(
        open_store,
    )

    assert askar_profile.opened == open_store


@pytest.mark.asyncio
async def test_init_multi_ledger(open_store):
    context = InjectionContext(
        settings={
            "ledger.ledger_config_list": [
                {
                    "id": "BCovrinDev",
                    "is_production": True,
                    "is_write": True,
                    "endorser_did": "9QPa6tHvBHttLg6U4xvviv",
                    "endorser_alias": "endorser_dev",
                    "genesis_transactions": mock.MagicMock(),
                },
                {
                    "id": "SovrinStagingNet",
                    "is_production": False,
                    "genesis_transactions": mock.MagicMock(),
                },
            ]
        }
    )
    askar_profile = AskarProfile(
        open_store,
        context=context,
    )

    assert askar_profile.opened == open_store
    assert askar_profile.settings["endorser.endorser_alias"] == "endorser_dev"
    assert (
        askar_profile.settings["endorser.endorser_public_did"] == "9QPa6tHvBHttLg6U4xvviv"
    )
    assert (askar_profile.inject_or(BaseLedger)).pool_name == "BCovrinDev"


@pytest.mark.asyncio
async def test_remove_success(open_store):
    openStore = open_store
    context = InjectionContext()
    profile_id = "profile_id"
    context.settings = {
        "multitenant.wallet_type": "single-wallet-askar",
        "wallet.askar_profile": profile_id,
        "ledger.genesis_transactions": mock.MagicMock(),
    }
    askar_profile = AskarProfile(openStore, context, profile_id=profile_id)
    remove_profile_stub = asyncio.Future()
    remove_profile_stub.set_result(True)
    openStore.store.remove_profile.return_value = remove_profile_stub

    await askar_profile.remove()

    openStore.store.remove_profile.assert_called_once_with(profile_id)


@pytest.mark.asyncio
async def test_remove_profile_not_removed_if_wallet_type_not_askar_profile(open_store):
    openStore = open_store
    context = InjectionContext()
    context.settings = {"multitenant.wallet_type": "basic"}
    askar_profile = AskarProfile(openStore, context)

    await askar_profile.remove()

    openStore.store.remove_profile.assert_not_called()


@pytest.mark.asyncio
async def test_profile_manager_transaction():
    profile = "profileId"

    with mock.patch("acapy_agent.askar.profile.AskarProfile") as AskarProfile:
        askar_profile = AskarProfile(None, True, profile_id=profile)
        askar_profile.profile_id = profile
        askar_profile_transaction = mock.MagicMock()
        askar_profile.store.transaction.return_value = askar_profile_transaction

        transactionProfile = test_module.AskarProfileSession(askar_profile, True)

        assert transactionProfile._opener == askar_profile_transaction
        askar_profile.store.transaction.assert_called_once_with(profile)


@pytest.mark.asyncio
async def test_profile_manager_store():
    config = {
        "test": True,
    }
    context = InjectionContext(
        settings=config,
    )
    await AskarProfileManager().provision(
        context=context,
        config=config,
    )
    await AskarAnonProfileManager().provision(
        context=context,
        config=config,
    )



"""A Base handler class for all message handlers."""

import logging
from abc import ABC, abstractmethod

from ..core.error import BaseError
from .request_context import RequestContext
from .responder import BaseResponder


class HandlerException(BaseError):
    """Exception base class for generic handler errors."""


class BaseHandler(ABC):
    """Abstract base class for handlers."""

    def __init__(self) -> None:
        """Initialize a BaseHandler instance."""
        self._logger = logging.getLogger(__name__)

    @abstractmethod
    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Abstract method for handler logic.

        Args:
            context: Request context object
            responder: A responder object

        """



"""Response to keylist-query message."""

from typing import Optional, Sequence

from marshmallow import fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import KEYLIST, PROTOCOL_PACKAGE
from .inner.keylist_key import KeylistKeySchema
from .inner.keylist_query_paginate import KeylistQueryPaginate, KeylistQueryPaginateSchema

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.keylist_handler.KeylistHandler"


class Keylist(AgentMessage):
    """Class representing a keylist-query response."""

    class Meta:
        """Metadata for a keylist query response."""

        handler_class = HANDLER_CLASS
        message_type = KEYLIST
        schema_class = "KeylistSchema"

    def __init__(
        self,
        *,
        keys: Sequence[str] = None,
        pagination: Optional[KeylistQueryPaginate] = None,
        **kwargs,
    ):
        """Initialize keylist query response object.

        Args:
            keys: Found keys by requested query
            pagination: Pagination rules
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.keys = list(keys) if keys else []
        self.pagination = pagination


class KeylistSchema(AgentMessageSchema):
    """Keylist query response schema class."""

    class Meta:
        """Keylist query response schema metadata."""

        model_class = Keylist

    pagination = fields.Nested(
        KeylistQueryPaginateSchema(),
        required=False,
        metadata={"description": "List of update rules"},
    )
    keys = fields.List(
        fields.Nested(KeylistKeySchema()), metadata={"description": "Keys"}
    )



"""Profile utilities."""

import json

from aiohttp import web

from ..anoncreds.error_messages import ANONCREDS_PROFILE_REQUIRED_MSG
from ..askar.profile_anon import AskarAnonCredsProfile
from ..core.profile import Profile
from ..multitenant.manager import MultitenantManager
from ..multitenant.single_wallet_askar_manager import SingleWalletAskarMultitenantManager
from ..storage.base import BaseStorageSearch
from ..wallet.models.wallet_record import WalletRecord


def is_anoncreds_profile_raise_web_exception(profile: Profile) -> None:
    """Raise a web exception when the supplied profile is anoncreds."""
    if isinstance(profile, AskarAnonCredsProfile):
        raise web.HTTPForbidden(reason="Interface not supported for an anoncreds profile")


def is_not_anoncreds_profile_raise_web_exception(profile: Profile) -> None:
    """Raise a web exception when the supplied profile is anoncreds."""
    if not isinstance(profile, AskarAnonCredsProfile):
        raise web.HTTPForbidden(reason=ANONCREDS_PROFILE_REQUIRED_MSG)


async def get_subwallet_profiles_from_storage(root_profile: Profile) -> list[Profile]:
    """Get subwallet profiles from storage."""
    subwallet_profiles = []
    base_storage_search = root_profile.inject(BaseStorageSearch)
    search_session = base_storage_search.search_records(
        type_filter=WalletRecord.RECORD_TYPE, page_size=10
    )
    if (
        root_profile.context.settings.get("multitenant.wallet_type")
        == "single-wallet-askar"
    ):
        manager = SingleWalletAskarMultitenantManager(root_profile)
    else:
        manager = MultitenantManager(root_profile)
    while search_session._done is False:
        wallet_storage_records = await search_session.fetch()
        for wallet_storage_record in wallet_storage_records:
            wallet_record = WalletRecord.from_storage(
                wallet_storage_record.id,
                json.loads(wallet_storage_record.value),
            )
            subwallet_profiles.append(
                await manager.get_wallet_profile(
                    base_context=root_profile.context,
                    wallet_record=wallet_record,
                )
            )
    return subwallet_profiles



"""Protocol controller for present proof v2_0."""

from typing import Sequence

VERIFY_VC = "aries.vc.verify"


class Controller:
    """Present proof v2_0 protocol controller."""

    def __init__(self, protocol: str):
        """Initialize the controller."""

    def determine_goal_codes(self) -> Sequence[str]:
        """Return defined goal_codes."""
        return [VERIFY_VC]



"""Validator methods to check for properties without a context."""

from typing import Optional, Sequence, Tuple, Union

from pyld import jsonld

from .document_loader import DocumentLoaderMethod


def diff_dict_keys(
    full: dict,
    with_missing: dict,
    prefix: Optional[str] = None,
    *,
    document_loader: DocumentLoaderMethod,
    context,
) -> Sequence[str]:
    """Get the difference in dict keys between full and with_missing.

    Checks recursively

    Args:
        full (dict): The full dict with all keys present
        with_missing (dict): The dict with possibly keys missing
        prefix (str, optional): The prefix. Mostly used for internal recursion.
        document_loader (DocumentLoaderMethod): The document loader method
        context: The context of the document

    Returns:
        Sequence[str]: List of missing property names in with_missing

    """

    def _normalize(
        full: Union[dict, list], with_missing: Union[dict, list]
    ) -> Tuple[Union[dict, list], Union[dict, list]]:
        full_type = type(full)
        with_missing_type = type(with_missing)

        if full_type == with_missing_type:
            return (full, with_missing)

        # First type is a list. Return first item if len is 1
        if full_type is list and with_missing_type is not list:
            return (full, [with_missing])

    missing = []

    # Loop trough all key/value pairs of the full document
    for key, value in full.items():
        key_in_with_missing = key

        # skip json-ld keywords
        if jsonld._is_keyword(key):
            continue

        _prefix = f"{prefix}.{key}" if prefix else key
        # If the key is not present in the with_missing dict, add it to the list
        if key not in with_missing:
            # The key could change for compaction. We check for the expanded version
            doc = {"@context": context, key: value}

            if full.get("type"):
                doc["type"] = full.get("type")
            elif full.get("@type"):
                doc["@type"] = full.get("@type")

            expanded = jsonld.expand(
                doc,
                {"documentLoader": document_loader},
            )

            if len(expanded) > 0:
                expanded = expanded[0]
                expanded.pop("@context", None)
                expanded.pop("@type", None)

                if len(expanded) == 1 and list(expanded.keys())[0] in with_missing:
                    key_in_with_missing = list(expanded.keys())[0]
                else:
                    missing.append(_prefix)
                    continue
            else:
                missing.append(_prefix)
                continue

        # If the key is present, but is a dict itself, recursively check nested keys
        if isinstance(value, dict):
            missing.extend(
                diff_dict_keys(
                    value,
                    with_missing.get(key_in_with_missing),
                    prefix=_prefix,
                    document_loader=document_loader,
                    context=context,
                )
            )

        # If the key is present, but is a list, recursively check nested keys for entries
        elif isinstance(value, list):
            value, value_with_missing = _normalize(
                value, with_missing.get(key_in_with_missing)
            )

            for i in range(len(value)):
                nested_value = value[i]

                # Only check for nested list or dict. We're not checking for string values
                if isinstance(nested_value, (dict, list)):
                    __prefix = f"{_prefix}[{i}]"

                    nested_value, nested_with_missing = _normalize(
                        nested_value, value_with_missing[i]
                    )
                    missing.extend(
                        diff_dict_keys(
                            nested_value,
                            nested_with_missing,
                            prefix=__prefix,
                            document_loader=document_loader,
                            context=context,
                        )
                    )

    return missing


def get_properties_without_context(
    document: dict, document_loader: DocumentLoaderMethod
) -> Sequence[str]:
    """Get the properties from document that don't have an context definition."""
    # FIXME: this doesn't work with nested @context structures...
    if "verifiableCredential" in document:
        return []

    document = document.copy()

    # Removes unknown keys from object
    compact = jsonld.compact(
        document,
        document["@context"],
        {"documentLoader": document_loader},
    )

    missing = diff_dict_keys(
        document, compact, document_loader=document_loader, context=document["@context"]
    )

    return missing



"""Represents a generic problem report message."""

from typing import Mapping, Optional, Sequence

from marshmallow import EXCLUDE, ValidationError, fields, validate, validates_schema

from ....messaging.agent_message import AgentMessage, AgentMessageSchema
from ....messaging.valid import RFC3339_DATETIME_EXAMPLE, RFC3339_DATETIME_VALIDATE
from .message_types import PROBLEM_REPORT, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handler.ProblemReportHandler"


class ProblemReport(AgentMessage):
    """Base class representing a generic problem report message."""

    class Meta:
        """Problem report metadata."""

        handler_class = HANDLER_CLASS
        message_type = PROBLEM_REPORT
        schema_class = "ProblemReportSchema"

    def __init__(
        self,
        *,
        description: Mapping[str, str] = None,
        problem_items: Sequence[Mapping[str, str]] = None,
        who_retries: Optional[str] = None,
        fix_hint: Mapping[str, str] = None,
        impact: Optional[str] = None,
        where: Optional[str] = None,
        noticed_time: Optional[str] = None,
        tracking_uri: Optional[str] = None,
        escalation_uri: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a ProblemReport message instance.

        Args:
            description: Human-readable, localized string(s) explaining the problem
            problem_items: List of problem items
            who_retries: you | me | both | none
            fix_hint: Dictionary of localized hints
            impact: message | thread | connection
            where: you | me | other (cloud | edge | wire | agency ..)
            noticed_time: Datetime when the problem was noticed
            tracking_uri: URI for tracking the problem
            escalation_uri: URI for escalating the problem
            kwargs: Additional keyword arguments for message
        """
        super().__init__(**kwargs)
        self.description = description if description else None
        self.problem_items = problem_items if problem_items else None
        self.who_retries = who_retries
        self.fix_hint = dict(fix_hint) if fix_hint else None
        self.impact = impact
        self.where = where
        self.noticed_time = noticed_time
        self.tracking_uri = tracking_uri
        self.escalation_uri = escalation_uri


class ProblemReportSchema(AgentMessageSchema):
    """Schema for ProblemReport base class."""

    class Meta:
        """Problem report schema metadata."""

        model_class = ProblemReport
        unknown = EXCLUDE

    description = fields.Dict(
        keys=fields.Str(metadata={"description": "Locale or 'code'", "example": "en-US"}),
        values=fields.Str(metadata={"description": "Problem description or error code"}),
        required=False,
        metadata={"description": "Human-readable localized problem descriptions"},
    )
    problem_items = fields.List(
        fields.Dict(
            keys=fields.Str(metadata={"description": "Problematic parameter or item"}),
            values=fields.Str(metadata={"description": "Problem text/number/value"}),
            metadata={"description": "Problem item"},
        ),
        required=False,
        metadata={"description": "List of problem items"},
    )
    who_retries = fields.Str(
        required=False,
        validate=validate.OneOf(["you", "me", "both", "none"]),
        metadata={
            "description": "Party to retry: you, me, both, none",
            "example": "you",
        },
    )
    fix_hint = fields.Dict(
        keys=fields.Str(metadata={"description": "Locale", "example": "en-US"}),
        values=fields.Str(
            metadata={
                "description": "Localized message",
                "example": "Synchronize time to NTP",
            }
        ),
        required=False,
        metadata={
            "description": "Human-readable localized suggestions how to fix problem"
        },
    )
    impact = fields.Str(
        required=False,
        validate=validate.OneOf(["message", "thread", "connection"]),
        metadata={
            "description": (
                "Breadth of impact of problem: message, thread, or connection"
            ),
            "example": "thread",
        },
    )
    where = fields.Str(
        required=False,
        validate=validate.Regexp("(you)|(me)|(other) - .+"),
        metadata={
            "description": "Where the error occurred, from reporter perspective",
            "example": "you - agency",
        },
    )
    noticed_time = fields.Str(
        data_key="noticed_time",
        required=False,
        validate=RFC3339_DATETIME_VALIDATE,
        metadata={
            "description": (
                "Problem detection time, precision at least day up to millisecond"
            ),
            "example": RFC3339_DATETIME_EXAMPLE,
        },
    )
    tracking_uri = fields.Str(
        required=False,
        metadata={
            "description": "URI allowing recipient to track error status",
            "example": "http://myservice.com/status",
        },
    )
    escalation_uri = fields.Str(
        required=False,
        metadata={
            "description": "URI to supply additional help",
            "example": "mailto://help.desk@myservice.com",
        },
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data has neither indy nor ld_proof

        """
        if not data.get("description", {}).get("code"):
            raise ValidationError("Field description.code must have a value")



"""Test Problem Report Message."""

from unittest import TestCase, mock

import pytest

from ......messaging.models.base import BaseModelError
from .. import problem_report as test_module
from ..problem_report import (
    CMProblemReport,
    CMProblemReportSchema,
    ProblemReportReason,
    ValidationError,
)


class TestCMProblemReportMessage(TestCase):
    """Test problem report."""

    def test_make_model(self):
        """Make problem report model."""
        data = CMProblemReport(
            description={
                "en": "Insufficient credit",
                "code": ProblemReportReason.MEDIATION_NOT_GRANTED.value,
            }
        ).serialize()
        model_instance = CMProblemReport.deserialize(data)
        assert isinstance(model_instance, CMProblemReport)

        model_instance.description["code"] = "extraneous code"
        with pytest.raises(BaseModelError):
            CMProblemReport.deserialize(model_instance)

    def test_validate_x(self):
        """Exercise validation requirements."""
        schema = CMProblemReportSchema()
        with pytest.raises(ValidationError):
            schema.validate_fields({})

    def test_validate_and_logger(self):
        """Capture ValidationError and Logs."""
        data = CMProblemReport(
            description={
                "en": "Insufficient credit",
                "code": "invalid_code",
            },
        ).serialize()
        with mock.patch.object(test_module, "LOGGER", autospec=True) as mock_logger:
            CMProblemReportSchema().validate_fields(data)
        mock_logger.warning.assert_called_once()
